# Copyright 2021 MosaicML. All Rights Reserved.

import logging
import tempfile
from dataclasses import dataclass
from os.path import join
from typing import List

import yahp as hp

from composer.core.types import Batch
from composer.datasets.hparams import DataloaderSpec, DatasetHparams

log = logging.getLogger(__name__)


def _split_dict_fn(batch: Batch, n_microbatches: int) -> List[Batch]:
    if isinstance(batch, dict):
        chunked = {k: v.chunk(n_microbatches) for k, v in batch.items()}
        num_chunks = len(list(chunked.values())[0])
        return [{k: v[idx] for k, v in chunked.items()} for idx in range(num_chunks)]
    else:
        raise ValueError(f'Expect batch from dataloader to be of type Dict[str, Tensor], but got {type(batch)}')


@dataclass
class LMDatasetHparams(DatasetHparams):
    """
    Defines a generic dataset class for autoregressive language models.
    """

    datadir: List[str] = hp.required("Path to the Huggingface Datasets directory.")
    split: str = hp.required("Whether to use 'train', 'validation' or 'test' split.")
    tokenizer_name: str = hp.required("The name of the tokenizer to preprocess text with.")
    num_tokens: int = hp.optional(doc='If desired, the number of tokens to truncate the dataset to.', default=0)
    seed: int = hp.optional("Which seed to use to generate train and validation splits.", default=5)
    subsample_ratio: float = hp.optional(default=1.0, doc='If desired, the percentage of the dataset to use.')
    train_sequence_length: int = hp.optional(
        default=1024, doc='Optionally, the ability to set a custom sequence length for the training dataset.')
    val_sequence_length: int = hp.optional(
        default=1024, doc='Optionally, the ability to set a custom sequence length for the validation dataset.')
    shuffle: bool = hp.optional("Whether to shuffle the dataset for each epoch.", default=True)
    drop_last: bool = hp.optional("Whether to drop the last samples for the last batch.", default=False)

    def initialize_object(self) -> DataloaderSpec:
        try:
            import datasets
            import transformers
        except ImportError:
            raise ImportError('huggingface transformers and datasets are not installed. '
                              'Please install with `pip install mosaicml-composer[nlp]`')
        self.tokenizer = transformers.AutoTokenizer.from_pretrained(self.tokenizer_name)  #type: ignore (thirdparty)
        self.config = transformers.AutoConfig.from_pretrained(self.tokenizer_name)  #type: ignore (thirdparty)
        lm_datasets = [datasets.load_from_disk(i) for i in self.datadir]  #type: ignore (thirdparty)

        # TODO (Moin): this re-loads a large dataset into memory three times -- can the REng team permit
        # returning a dataloader for a particular split?
        if self.split not in ['train', 'validation', 'test']:
            raise ValueError("The dataset split must be one of 'train', 'validation', or 'test'.")

        # merge the dataset to re-sample from
        merged_dataset = [[d[self.split]] for d in lm_datasets]
        # flatten merged_dataset
        merged_dataset = [item for sublist in merged_dataset for item in sublist]
        lm_datasets = datasets.concatenate_datasets(merged_dataset)  #type: ignore (thirdparty)

        # generate a cache file name so the training and validation set use the same split
        indices_cache_file_name = join(tempfile.gettempdir(), f"{self.seed}.indicies")

        # shuffle the dataset
        lm_datasets = lm_datasets.shuffle(indices_cache_file_name=indices_cache_file_name, seed=self.seed)

        if self.num_tokens > 0 and self.subsample_ratio < 1.0:
            raise Exception("Must specify one of num_tokens OR subsample_ratio, cannot specify both.")

        total_num_samples = len(lm_datasets)
        tokens_per_sample = len(lm_datasets[0]['input_ids'])
        total_num_tokens = total_num_samples * tokens_per_sample

        # truncate the dataset to a specified size
        num_samples = total_num_samples
        if self.num_tokens > 0:
            assert self.num_tokens <= total_num_tokens, f"Requested {self.num_tokens} tokens must be <= total_num_tokens={total_num_tokens}"
            assert self.num_tokens % tokens_per_sample == 0, f"Requested {self.num_tokens} tokens is not divisible by tokens_per_sample={tokens_per_sample}"
            num_samples = self.num_tokens // tokens_per_sample
            self.subsample_ratio = num_samples / total_num_samples
        elif self.subsample_ratio < 1.0:
            num_samples = round(total_num_samples * self.subsample_ratio)
            self.num_tokens = num_samples * tokens_per_sample
        else:
            log.warning("No subsampling going on!")

        lm_datasets = lm_datasets.select(range(num_samples))
        log.info(f"LM datasets: {lm_datasets}")
        log.info(f"Subsample ratio: {self.subsample_ratio}")
        log.info(f"Total number of samples: {num_samples:e}")
        log.info(f"Total number of tokens: {self.num_tokens:e}")
        self.dataset = lm_datasets

        self.data_collator = transformers.default_data_collator

        return DataloaderSpec(
            dataset=self.dataset,  #type: ignore (thirdparty)
            collate_fn=self.data_collator,
            shuffle=self.shuffle,
            drop_last=self.drop_last,
            split_fn=_split_dict_fn)



# Copyright 2021 MosaicML. All Rights Reserved.

import pytest
import torch

from composer.algorithms import MixUpHparams
from composer.algorithms.mixup.mixup import gen_interpolation_lambda, mixup_batch
from composer.core.types import Event
from composer.models.base import MosaicClassifier
from composer.trainer.trainer_hparams import TrainerHparams
from tests.utils.trainer_fit import train_model


# (N, C, d1, d2, n_classes)
@pytest.fixture(params=[(7, 11, 3, 5, 10)])
def fake_data(request):
    # Generate some fake data
    N, C, d1, d2, n_classes = request.param
    torch.manual_seed(0)
    x_fake = torch.randn(N, C, d1, d2)
    y_fake = torch.randint(n_classes, size=(N,))
    indices = torch.randperm(N)
    return x_fake, y_fake, indices


def validate_mixup_batch(x, y, indices, x_mix, int_lamb):
    # Create shuffled version of y_fake for reference checking
    y_perm = y[indices]

    # Explicitly check that the batches and labels have been mixed correctly.
    for i in range(x.size(0)):  # Grab N
        j = indices[i]
        # Check the input data
        x_mix_test = (1 - int_lamb) * x[i] + int_lamb * x[j]
        torch.testing.assert_allclose(x_mix_test, x_mix[i])
        # Check the label
        perm_label = y[j]
        torch.testing.assert_allclose(perm_label, y_perm[i])


@pytest.mark.parametrize('alpha', [.2, 1])
class TestMixUp:

    def test_mixup_batch(self, fake_data, alpha):
        # Generate fake data
        x_fake, y_fake, indices = fake_data

        # Get interpolation lambda based on alpha hparam
        interpolation_lambda = gen_interpolation_lambda(alpha)

        # Apply mixup
        x_mix, y_mix, _ = mixup_batch(
            x=x_fake,
            y=y_fake,
            interpolation_lambda=interpolation_lambda,
            n_classes=x_fake.size(1),  # Grab C
            indices=indices)

        # Validate results
        validate_mixup_batch(x_fake, y_fake, indices, x_mix, interpolation_lambda)

    def test_mixup_algorithm(self, fake_data, alpha, dummy_state, dummy_logger):
        # Generate fake data
        x_fake, y_fake, _ = fake_data

        algorithm = MixUpHparams(alpha=alpha).initialize_object()
        state = dummy_state
        state.model = MosaicClassifier
        state.model.num_classes = x_fake.size(1)  # Grab C
        state.batch = (x_fake, y_fake)

        algorithm.apply(Event.INIT, state, dummy_logger)
        # Apply algo, use test hooks to specify indices and override internally generated interpolation lambda for testability
        algorithm.apply(Event.AFTER_DATALOADER, state, dummy_logger)

        x, _ = state.batch
        # Use algorithm generated indices and interpolation_lambda for validation
        validate_mixup_batch(x_fake, y_fake, algorithm.indices, x, algorithm.interpolation_lambda)


@pytest.mark.xfail
def test_mixup_nclasses(dummy_state, dummy_logger):
    algorithm = MixUpHparams(alpha=0.2).initialize_object()
    state = dummy_state
    state.model = MosaicClassifier
    state.model.num_classes = None  # This should flag AttributeError

    algorithm.apply(Event.AFTER_DATALOADER, state, dummy_logger)


@pytest.mark.run_long
@pytest.mark.timeout(90)
def test_mixup_trains(mosaic_trainer_hparams: TrainerHparams):
    mosaic_trainer_hparams.algorithms = [MixUpHparams(alpha=0.2)]
    train_model(mosaic_trainer_hparams)



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.models.efficientnetb0.efficientnetb0_hparams import EfficientNetB0Hparams as EfficientNetB0Hparams
from composer.models.efficientnetb0.model import EfficientNetB0 as EfficientNetB0

_task = 'Image Classification'
_dataset = 'ImageNet'
_name = 'EfficientNet-B0'
_quality = ''
_metric = 'Top-1 Accuracy'
_ttt = '?'
_hparams = 'hparams.yaml'



# Copyright 2021 MosaicML. All Rights Reserved.

import random

import numpy as np
import torch


def get_random_seed() -> int:
    seed = int(torch.empty((), dtype=torch.int64).random_(to=2**32).item())
    return seed


def seed_all(seed: int):
    """
    Seed all rng objects

    Args:
        seed (int): random seed
        device (Optional[Device]): the
    """

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    # torch.manual_seed may call manual_seed_all but calling it again here
    # to make sure it gets called at least once
    torch.cuda.manual_seed_all(seed)



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import functools
import logging
from dataclasses import asdict, dataclass
from typing import Optional

import numpy as np
import torch
import yahp as hp

from composer.algorithms import AlgorithmHparams
from composer.algorithms.blurpool.blurpool_layers import BlurConv2d, BlurMaxPool2d
from composer.core import Algorithm, Event, Logger, State, surgery

log = logging.getLogger(__name__)


def _log_surgery_result(model: torch.nn.Module):
    num_blurpool_layers = surgery.count_module_instances(model, BlurMaxPool2d)
    num_blurconv_layers = surgery.count_module_instances(model, BlurConv2d)
    log.info(f'Applied BlurPool to model {model.__class__.__name__}. '
             f'Model now has {num_blurpool_layers} BlurMaxPool2d '
             f'and {num_blurconv_layers} BlurConv2D layers.')


def apply_blurpool(model: torch.nn.Module,
                   replace_convs: bool = True,
                   replace_maxpools: bool = True,
                   blur_first: bool = True) -> None:
    """Applies BlurPool algorithm to the provided model. Performs an in-place
    replacement of eligible convolution and pooling layers.

    Args:
        model (torch.nn.Module): model to transform
        replace_convs (bool): replace eligible Conv2D with BlurConv2d. Default: True.
        replace_maxpools (bool): replace eligible MaxPool2d with BlurMaxPool2d. Default: True.
        blur_first (bool): for replace_convs, blur input before conv. Default: True
    """
    transforms = {}
    if replace_maxpools:
        transforms[torch.nn.MaxPool2d] = BlurMaxPool2d.from_maxpool2d
    if replace_convs:
        transforms[torch.nn.Conv2d] = functools.partial(
            _maybe_replace_strided_conv2d,
            blur_first=blur_first,
        )
    surgery.replace_module_classes(model, policies=transforms)
    _log_surgery_result(model)


@dataclass
class BlurPoolHparams(AlgorithmHparams):

    replace_convs: bool = hp.optional('Replace Conv2d with BlurConv2d if stride > 1', default=True)
    replace_maxpools: bool = hp.optional('Replace MaxPool2d with BlurMaxPool2d', default=True)
    blur_first: bool = hp.optional('Blur input before Conv', default=True)

    def initialize_object(self) -> "BlurPool":
        return BlurPool(**asdict(self))


def _maybe_replace_strided_conv2d(module: torch.nn.Conv2d, module_index: int, blur_first: bool):
    if (np.max(module.stride) > 1 and module.in_channels >= 16):
        return BlurConv2d.from_conv2d(module, module_index, blur_first=blur_first)
    return None


class BlurPool(Algorithm):
    """Algorithm to apply BlurPool to the model. Runs on Event.INIT. This algorithm should
    be applied before the model has been moved to devices.

    Args:
        replace_convs (bool): replace eligible Conv2D with BlurConv2d. Default: True.
        replace_maxpools (bool): replace eligible MaxPool2d with BlurMaxPool2d. Default: True.
        blur_first (bool): for replace_convs, blur input before conv. Default: True
    """

    def __init__(self, replace_convs: bool, replace_maxpools: bool, blur_first: bool) -> None:
        self.hparams = BlurPoolHparams(
            replace_convs=replace_convs,
            replace_maxpools=replace_maxpools,
            blur_first=blur_first,
        )

        if self.hparams.replace_maxpools is False and \
             self.hparams.replace_convs is False:
            log.warning('Both replace_maxpool and replace_convs set to false '
                        'BlurPool will not be modifying the model.')

    def match(self, event: Event, state: State) -> bool:
        """ Runs on Event.INIT
        """
        return event == Event.INIT

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        """ Applies BlurPool
        """
        assert state.model is not None

        apply_blurpool(state.model, **asdict(self.hparams))
        self._log_results(event, state, logger)

    def _log_results(self, event: Event, state: State, logger: Logger) -> None:
        """ Logs the result of BlurPool application, including the number
        of layers that have been replaced.
        """
        assert state.model is not None

        num_blurpool_layers = surgery.count_module_instances(state.model, BlurMaxPool2d)
        num_blurconv_layers = surgery.count_module_instances(state.model, BlurConv2d)

        # python logger
        # TODO: standardize algorithm logging
        log.info(f'Applied BlurPool to model {state.model.__class__.__name__} '
                 f'with replace_maxpools={self.hparams.replace_maxpools}, '
                 f'replace_convs={self.hparams.replace_convs}. '
                 f'Model now has {num_blurpool_layers} BlurMaxPool2d '
                 f'and {num_blurconv_layers} BlurConv2D layers.')

        logger.metric_fit({
            'blurpool/num_blurpool_layers': num_blurpool_layers,
            'blurpool/num_blurconv_layers': num_blurconv_layers,
        })



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.stochastic_depth.sample_stochastic_layers import \
    SampleStochasticBottleneck as SampleStochasticBottleneck
from composer.algorithms.stochastic_depth.stochastic_depth import StochasticDepth as StochasticDepth
from composer.algorithms.stochastic_depth.stochastic_depth import StochasticDepthHparams as StochasticDepthHparams
from composer.algorithms.stochastic_depth.stochastic_depth import apply_stochastic_depth as apply_stochastic_depth
from composer.algorithms.stochastic_depth.stochastic_layers import StochasticBottleneck as StochasticBottleneck

_name = 'Stochastic Depth'
_class_name = 'StochasticDepth'
_functional = 'apply_stochastic_depth'
_tldr = 'Replaces a specified layer with a stochastic verion that randomly drops the layer or samples during training'
_attribution = '(Huang et al, 2016)'
_link = 'https://arxiv.org/abs/1603.09382'
_method_card = ''



# Copyright 2021 MosaicML. All Rights Reserved.

import collections.abc
from unittest.mock import MagicMock

import pytest

from composer.callbacks import SpeedMonitorHparams
from composer.trainer import TrainerHparams


@pytest.mark.timeout(60)
@pytest.mark.run_long
def test_speed_monitor(mosaic_trainer_hparams: TrainerHparams):
    speed_monitor_hparams = SpeedMonitorHparams(window_size=2)
    mosaic_trainer_hparams.callbacks.append(speed_monitor_hparams)
    mosaic_trainer_hparams.grad_accum = 1
    mosaic_trainer_hparams.ddp.fork_rank_0 = False
    mosaic_trainer_hparams.total_batch_size = 10
    mosaic_trainer_hparams.max_epochs = 2
    trainer = mosaic_trainer_hparams.initialize_object()
    log_destination = MagicMock()
    log_destination.will_log.return_value = True
    trainer.logger.backends = [log_destination]
    trainer.fit()

    throughput_epoch_calls = 0
    wall_clock_train_calls = 0
    throughput_step_calls = 0
    for call_ in log_destination.log_metric.mock_calls:
        metrics = call_[1][3]
        if "throughput/step" in metrics:
            throughput_step_calls += 1
        if "throughput/epoch" in metrics:
            throughput_epoch_calls += 1
        if 'wall_clock_train' in metrics:
            wall_clock_train_calls += 1

    assert isinstance(trainer.state.train_dataloader, collections.abc.Sized)
    expected_step_calls = (len(trainer.state.train_dataloader) -
                           speed_monitor_hparams.window_size) * mosaic_trainer_hparams.max_epochs
    assert throughput_step_calls == expected_step_calls
    assert throughput_epoch_calls == mosaic_trainer_hparams.max_epochs
    assert wall_clock_train_calls == mosaic_trainer_hparams.max_epochs



# Copyright 2021 MosaicML. All Rights Reserved.

from unittest.mock import Mock

import pytest

from composer.core import Event, engine


@pytest.fixture
def always_match_algorithms():
    attrs = {'match.return_value': True}
    return [
        Mock(**{
            'match.return.value': True,
            'apply.return_value': n,  # return encodes order
        }) for n in range(5)
    ]


@pytest.fixture
def never_match_algorithms():
    attrs = {'match.return_value': False}
    return [Mock(**attrs) for _ in range(5)]


def run_event(event, state, algorithms, logger):
    runner = engine.Engine(state, algorithms, logger)
    return runner.run_event(event)


@pytest.mark.parametrize('event', list(Event))
class TestAlgorithms:

    def test_algorithms_always_called(self, event, dummy_state, always_match_algorithms, dummy_logger):
        _ = run_event(event, dummy_state, always_match_algorithms, dummy_logger)
        for algo in always_match_algorithms:
            algo.apply.assert_called_once()
            algo.match.assert_called_once()

    def test_algorithms_never_called(self, event, dummy_state, never_match_algorithms, dummy_logger):
        _ = run_event(event, dummy_state, never_match_algorithms, dummy_logger)
        for algo in never_match_algorithms:
            algo.apply.assert_not_called()
            algo.match.assert_called_once()

    def test_engine_trace_all(self, event, dummy_state, always_match_algorithms, dummy_logger):
        trace = run_event(event, dummy_state, always_match_algorithms, dummy_logger)

        assert all([tr.run for tr in trace.values()])

    def test_engine_trace_never(self, event, dummy_state, never_match_algorithms, dummy_logger):
        trace = run_event(event, dummy_state, never_match_algorithms, dummy_logger)

        assert all([tr.run is False for tr in trace.values()])


@pytest.mark.parametrize(
    'event',
    [
        Event.TRAINING_START,
        Event.BEFORE_LOSS,
        Event.BEFORE_BACKWARD,
        Event.TRAINING_END,  # no before prefix, so run in normal order
    ])
def test_engine_lifo_first_in(event, dummy_state, always_match_algorithms, dummy_logger):
    trace = run_event(event, dummy_state, always_match_algorithms, dummy_logger)
    order = [tr.order for tr in trace.values()]
    expected_order = [tr.exit_code for tr in trace.values()]  # use exit_code to uniquely label algos

    assert order == expected_order


@pytest.mark.parametrize('event', [
    Event.AFTER_LOSS,
    Event.AFTER_BACKWARD,
])
def test_engine_lifo_last_out(event, dummy_state, always_match_algorithms, dummy_logger):
    trace = run_event(event, dummy_state, always_match_algorithms, dummy_logger)
    order = [tr.order for tr in trace.values()]
    expected_order = list(reversed([tr.exit_code for tr in trace.values()]))

    assert order == expected_order



# Copyright 2021 MosaicML. All Rights Reserved.

from dataclasses import dataclass

import yahp as hp
from torchvision import datasets, transforms

from composer.datasets.hparams import DataloaderSpec, DatasetHparams


@dataclass
class MNISTDatasetHparams(DatasetHparams):
    is_train: bool = hp.required("whether to load the training or validation dataset")
    datadir: str = hp.required("data directory")
    download: bool = hp.required("whether to download the dataset, if needed")
    drop_last: bool = hp.optional("Whether to drop the last samples for the last batch", default=True)
    shuffle: bool = hp.optional("Whether to shuffle the dataset for each epoch", default=True)

    def initialize_object(self) -> DataloaderSpec:
        transform = transforms.Compose([transforms.ToTensor()])
        dataset = datasets.MNIST(
            self.datadir,
            train=self.is_train,
            download=self.download,
            transform=transform,
        )
        return DataloaderSpec(
            dataset=dataset,
            drop_last=self.drop_last,
            shuffle=self.shuffle,
        )



# Copyright 2021 MosaicML. All Rights Reserved.

from typing import Callable

import pytest
import torch

from composer.algorithms.selective_backprop import SelectiveBackprop, SelectiveBackpropHparams
from composer.algorithms.selective_backprop.selective_backprop import do_selective_backprop, selective_backprop
from composer.core import Event
from composer.core.logging.logger import Logger
from composer.core.state import State
from composer.core.types import DataLoader
from composer.models import MosaicClassifier
from composer.trainer.trainer_hparams import TrainerHparams
from tests.utils.trainer_fit import train_model


@pytest.fixture
def N() -> int:
    """Batch size
    """
    return 16


@pytest.fixture
def D() -> int:
    """Input dimension
    """
    return 8


@pytest.fixture
def X(N: int, D: int) -> torch.Tensor:
    """2D input
    """
    torch.manual_seed(42)
    return torch.randn(N, D)


@pytest.fixture
def X3D(N: int, D: int) -> torch.Tensor:
    """3D input
    """
    torch.manual_seed(42)
    return torch.randn(N, D, D)


@pytest.fixture
def X5D(N: int, D: int) -> torch.Tensor:
    """5D input
    """
    torch.manual_seed(42)
    return torch.randn(N, D, D, D, D)


@pytest.fixture
def Ximage(N: int) -> torch.Tensor:
    """4D image input
    """
    torch.manual_seed(42)
    return torch.randn(N, 3, 32, 32)


@pytest.fixture
def y(N: int) -> torch.Tensor:
    """Target
    """
    torch.manual_seed(42)
    return torch.randint(2, (N,))


@pytest.fixture
def loss_fun() -> Callable:
    """Fake loss function
    """

    def loss(output, target, reduction="none"):
        #import pdb; pdb.set_trace()
        return torch.ones_like(target)

    return loss


@pytest.fixture
def loss_fun_tuple() -> Callable:
    """Fake loss function that requires a batch tuple
    """

    def loss(output, batch, reduction="none"):
        _, target = batch
        return torch.ones_like(target)

    return loss


@pytest.fixture
def bad_loss() -> Callable:
    """Fake loss function that will error
    """

    def loss(output, target):
        return 0

    return loss


@pytest.fixture
def model(X: torch.Tensor) -> torch.nn.Module:
    """Simple fake linear model
    """
    return torch.nn.Linear(X.shape[1], 1)


@pytest.fixture
def model3D(X3D: torch.Tensor) -> torch.nn.Module:
    """Simple fake linear model
    """
    return torch.nn.Sequential(torch.nn.AdaptiveAvgPool1d(1), torch.nn.Flatten(), torch.nn.Linear(X3D.shape[1], 1))


@pytest.fixture
def model5D(X5D: torch.Tensor) -> torch.nn.Module:
    """Simple fake linear model
    """
    return torch.nn.Sequential(torch.nn.AdaptiveAvgPool3d(1), torch.nn.Flatten(), torch.nn.Linear(X5D.shape[1], 1))


@pytest.fixture
def start() -> float:
    """start hparam
    """
    return 0.5


@pytest.fixture
def end() -> float:
    """end hparam
    """
    return 0.8


@pytest.fixture
def keep() -> float:
    """keep hparam
    """
    return 0.5


@pytest.fixture
def scale_factor() -> float:
    """scale_factor hparam
    """
    return 0.5


@pytest.fixture
def interrupt() -> int:
    """interrupt hparams
    """
    return 2


@pytest.fixture
def dummy_hparams(start: float, end: float, keep: float, scale_factor: float,
                  interrupt: int) -> SelectiveBackpropHparams:
    """Dummy algo hparams
    """
    return SelectiveBackpropHparams(start, end, keep, scale_factor, interrupt)


@pytest.fixture
def dummy_algorithm(dummy_hparams: SelectiveBackpropHparams) -> SelectiveBackprop:
    """Dummy algorithm
    """
    return dummy_hparams.initialize_object()


@pytest.fixture
def epoch() -> int:
    """Default epoch
    """
    return 5


@pytest.fixture
def batch() -> int:
    """Default batch
    """
    return 0


@pytest.fixture
def conv_model(Ximage: torch.Tensor, D: int) -> MosaicClassifier:
    """Dummy conv model
    """
    return MosaicClassifier(torch.nn.Conv2d(Ximage.shape[1], D, 3))


@pytest.fixture
def dummy_state_sb(dummy_state: State, dummy_train_dataloader: DataLoader, conv_model: MosaicClassifier,
                   loss_fun_tuple: Callable, epoch: int, batch: int) -> State:
    """Dummy state with required values set for Selective Backprop
    """

    dummy_state.train_dataloader = dummy_train_dataloader
    dummy_state.epoch = epoch
    dummy_state.step = epoch * len(dummy_train_dataloader) + batch
    dummy_state.model = conv_model
    dummy_state.model.module.loss = loss_fun_tuple

    return dummy_state


@pytest.mark.parametrize("epoch,batch,interrupt", [(10, 0, 0), (10, 0, 2), (10, 2, 2)])
def test_do_selective_backprop_true(epoch: int, batch: int, interrupt: int) -> None:
    """Test functional match when epoch is within interval
    """
    start = 5
    end = 15
    is_chosen = do_selective_backprop(epoch, batch, start, end, interrupt)
    assert is_chosen


@pytest.mark.parametrize("epoch,batch,interrupt", [(0, 0, 0), (20, 0, 0), (10, 1, 2)])
def test_do_selective_backprop_false(epoch: int, batch: int, interrupt: int) -> None:
    """Test functional doesn't match when epoch is outside of interval
    """
    start = 5
    end = 15
    is_chosen = do_selective_backprop(epoch, batch, start, end, interrupt)
    assert not is_chosen


@pytest.mark.parametrize("keep", [0.5])
@pytest.mark.parametrize("scale_factor", [0.5])
@pytest.mark.xfail()
def test_selective_output_shape_3D(X3D: torch.Tensor, y: torch.Tensor, model: torch.nn.Module, loss_fun: Callable,
                                   keep: float, scale_factor: float) -> None:
    """Test functional selection on 3D inputs
    """
    N, D, _ = X3D.shape

    X_scaled, y_scaled = selective_backprop(X3D, y, model, loss_fun, keep, scale_factor)
    assert X_scaled.shape == (int(N * keep), D, D)
    assert y_scaled.shape == (int(N * keep),)


@pytest.mark.parametrize("keep", [1, 0.5, 0.75])
@pytest.mark.parametrize("scale_factor", [1])
def test_selective_output_shape(X: torch.Tensor, y: torch.Tensor, model: torch.nn.Module, loss_fun: Callable,
                                keep: float, scale_factor: float) -> None:
    """Test functional selection on 2D inputs
    """
    N, D = X.shape

    X_scaled, y_scaled = selective_backprop(X, y, model, loss_fun, keep, scale_factor)
    assert X_scaled.shape == (int(N * keep), D)
    assert y_scaled.shape == (int(N * keep),)


@pytest.mark.parametrize("keep", [0.5, 0.75, 1])
@pytest.mark.parametrize("scale_factor", [0.5, 0.75])
def test_selective_output_shape_scaled(Ximage: torch.Tensor, y: torch.Tensor, conv_model: MosaicClassifier,
                                       loss_fun: Callable, keep: float, scale_factor: float) -> None:
    """Test functional selection on 4D inputs
    """
    N, C, H, W = Ximage.shape
    X_scaled, y_scaled = selective_backprop(Ximage, y, conv_model.module, loss_fun, keep, scale_factor)
    assert X_scaled.shape == (int(N * keep), C, H, W)
    assert y_scaled.shape == (int(N * keep),)


def test_selective_backprop_interp_dim_error(X: torch.Tensor, y: torch.Tensor, model: torch.nn.Module,
                                             loss_fun: Callable) -> None:
    """Ensure that ValueError is raised when input tensor can't be scaled
    """
    with pytest.raises(ValueError):
        X_scaled, y_scaled = selective_backprop(X, y, model, loss_fun, 1, 0.5)


def test_selective_backprop_bad_loss_error(X: torch.Tensor, y: torch.Tensor, model: torch.nn.Module,
                                           bad_loss: Callable) -> None:
    """Ensure that ValueError is raised when loss function doesn't have `reduction` kwarg
    """
    with pytest.raises(TypeError) as execinfo:
        X_scaled, y_scaled = selective_backprop(X, y, model, bad_loss, 1, 1)
    MATCH = "must take a keyword argument `reduction`."
    assert MATCH in str(execinfo.value)


@pytest.mark.parametrize("event", [Event.AFTER_DATALOADER])
@pytest.mark.parametrize("epoch,batch", [(5, 0), (7, 0), (5, 2)])
def test_match_correct(event: Event, dummy_algorithm: SelectiveBackprop, dummy_state_sb: State) -> None:
    """ Algo should match AFTER_DATALOADER in the right interval
    """
    dummy_state_sb.max_epochs = 10

    assert dummy_algorithm.match(event, dummy_state_sb)


@pytest.mark.parametrize("event,epoch,batch", [(Event.TRAINING_START, 5, 0), (Event.AFTER_DATALOADER, 0, 0),
                                               (Event.AFTER_DATALOADER, 5, 1)])
def test_match_incorrect(event: Event, dummy_algorithm: SelectiveBackprop, dummy_state_sb: State) -> None:
    """ Algo should NOT match TRAINING_START or the wrong interval
    """
    dummy_state_sb.max_epochs = 10

    assert not dummy_algorithm.match(event, dummy_state_sb)


@pytest.mark.parametrize("epoch,batch", [(5, 0)])
@pytest.mark.parametrize("keep", [0.5, 0.75, 1])
@pytest.mark.parametrize("scale_factor", [0.5, 1])
def test_apply(Ximage: torch.Tensor, y: torch.Tensor, dummy_algorithm: SelectiveBackprop, dummy_state_sb: State,
               dummy_logger: Logger, keep: float) -> None:
    """Test apply with image inputs gives the right output shape
    """
    N, C, H, W = Ximage.shape

    dummy_state_sb.max_epochs = 10
    dummy_state_sb.batch = (Ximage, y)
    dummy_algorithm.apply(Event.AFTER_DATALOADER, dummy_state_sb, dummy_logger)

    X_scaled, y_scaled = dummy_state_sb.batch
    assert X_scaled.shape == (int(N * keep), C, H, W)
    assert y_scaled.shape == (int(N * keep),)


@pytest.mark.run_long
@pytest.mark.timeout(90)
def test_selective_backprop_trains(mosaic_trainer_hparams: TrainerHparams):
    mosaic_trainer_hparams.algorithms = [
        SelectiveBackpropHparams(start=0.3, end=0.9, keep=0.75, scale_factor=0.5, interrupt=1)
    ]
    train_model(mosaic_trainer_hparams, max_epochs=6)



# Copyright 2021 MosaicML. All Rights Reserved.

import torch

from composer.models.resnets import Bottleneck


def _sample_bernoulli(probability: torch.Tensor,
                      device_id: int,
                      module_id: int,
                      num_modules: int,
                      generator: torch.Generator,
                      use_same_depth_across_gpus: bool,
                      use_same_gpu_seed: bool = False):
    """Gets a sample from a Bernoulli distribution. Provides functionality to
       have different seeds across GPUs and to have the same set of seeds across GPUs.
    """

    if use_same_gpu_seed:
        sample = torch.bernoulli(probability)
        return sample

    # Get a separate generator for each GPU
    rand_int = torch.randint(low=2**17, high=2**27, size=(1,), dtype=torch.long, generator=generator)
    gpu_seed = (rand_int * (device_id + 1))

    if use_same_depth_across_gpus:
        gpu_generator = torch.Generator().manual_seed(gpu_seed.item())  #type: ignore
        layer_seed = (torch.randperm(num_modules, generator=gpu_generator)[module_id] + 1) * rand_int
    else:
        layer_seed = (module_id + 1) * gpu_seed
    layer_generator = torch.Generator().manual_seed(layer_seed.item())  # type: ignore

    sample = torch.bernoulli(probability, generator=layer_generator)
    return sample


class StochasticBottleneck(Bottleneck):
    """Stochastic ResNet Bottleneck layer. This layer has a probability of skipping
       the transformation section of the layer and scales the transformation section
       output by (1 - drop probability) during inference.

       Args:
            drop_rate (float): Probability of dropping the layer. Must be
                                     between 0.0 and 1.0.
            module_id (int): The placement of the layer within a network e.g. 0
                             for the first layer in the network.
            module_count (int): The total number of layers of this type in the
                                network.
            use_same_gpu_seed (bool): Set to true to have the same layers dropped
                                      across GPUs when using multi-GPU training.
                                      Set to false to have each GPU drop a different
                                      set of layers.
            use_same_depth_across_gpus (bool): Set to true to have the same number
                                               of layers dropped across GPUs. Set
                                               to true when `drop_distribution`
                                               is 'uniform' and set to false for
                                               'linear'.

       """

    def __init__(self, drop_rate: float, module_id: int, module_count: int, use_same_gpu_seed: bool,
                 use_same_depth_across_gpus: bool, rand_generator: torch.Generator, **kwargs):
        super(StochasticBottleneck, self).__init__(**kwargs)
        self.drop_rate = torch.tensor(drop_rate)
        self.module_id = module_id
        self.module_count = module_count
        self.use_same_gpu_seed = use_same_gpu_seed
        self.use_same_depth_across_gpus = use_same_depth_across_gpus
        self.rand_generator = rand_generator

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x

        sample = _sample_bernoulli(probability=(1 - self.drop_rate),
                                   device_id=x.get_device(),
                                   module_id=self.module_id,
                                   num_modules=self.module_count,
                                   generator=self.rand_generator,
                                   use_same_gpu_seed=self.use_same_gpu_seed,
                                   use_same_depth_across_gpus=self.use_same_depth_across_gpus)

        if not self.training or sample:
            out = self.conv1(x)
            out = self.bn1(out)
            out = self.relu(out)

            out = self.conv2(out)
            out = self.bn2(out)
            out = self.relu(out)

            out = self.conv3(out)
            out = self.bn3(out)

            if not self.training:
                out = out * (1 - self.drop_rate)

            if self.downsample is not None:
                identity = self.downsample(x)

            out += identity
            out = self.relu(out)
        else:
            if self.downsample is not None:
                out = self.relu(self.downsample(x))
            else:
                out = identity
        return out

    @staticmethod
    def from_target_layer(module: Bottleneck,
                          module_index: int,
                          module_count: int,
                          drop_rate: float,
                          drop_distribution: str,
                          rand_generator: torch.Generator,
                          use_same_gpu_seed: bool = False):
        """Helper function to convert a ResNet bottleneck block into a stochastic block.
        """
        if drop_distribution == 'linear':
            drop_rate = ((module_index + 1) / module_count) * drop_rate
        use_same_depth_across_gpus = (drop_distribution == 'uniform')
        rand_generator = torch.Generator().manual_seed(
            rand_generator.initial_seed())  # copy the generator for each layer
        return StochasticBottleneck(drop_rate=drop_rate,
                                    module_id=module_index,
                                    module_count=module_count,
                                    use_same_depth_across_gpus=use_same_depth_across_gpus,
                                    use_same_gpu_seed=use_same_gpu_seed,
                                    rand_generator=rand_generator,
                                    inplanes=module.conv1.in_channels,
                                    planes=module.conv3.out_channels // module.expansion,
                                    stride=module.stride,
                                    downsample=module.downsample,
                                    groups=module.conv2.groups,
                                    dilation=module.conv2.dilation)



# Copyright 2021 MosaicML. All Rights Reserved.

"""
Test the blurpool algorithm. Primitives are tested in test_blurpool.py
"""
import itertools
from unittest.mock import MagicMock

import pytest
import torch

from composer.algorithms import BlurPool, BlurPoolHparams
from composer.algorithms.blurpool.blurpool_layers import BlurConv2d, BlurMaxPool2d
from composer.core import Event, State
from composer.core.types import Model, Precision
from tests.fixtures.dummy_fixtures import SimpleConvModel


@pytest.fixture
def state(simple_conv_model: Model):
    state = State(
        epoch=50,
        step=50,
        train_batch_size=100,
        eval_batch_size=100,
        grad_accum=1,
        max_epochs=100,
        model=simple_conv_model,
        precision=Precision.FP32,
    )
    return state


@pytest.fixture(params=itertools.product([True, False], [True, False], [True, False]))
def blurpool_instance(request) -> BlurPool:
    replace_conv, replace_pool, blur_first = request.param
    blurpool_hparams = BlurPoolHparams(
        replace_convs=replace_conv,
        replace_maxpools=replace_pool,
        blur_first=blur_first,
    )
    return blurpool_hparams.initialize_object()


@pytest.fixture
def dummy_logger():
    return MagicMock()


def test_blurconv(state, blurpool_instance, dummy_logger):
    blurpool_instance.apply(Event.INIT, state, dummy_logger)
    assert isinstance(state.model.module, SimpleConvModel)

    if blurpool_instance.hparams.replace_convs:
        assert type(state.model.module.conv1) is BlurConv2d
    else:
        assert type(state.model.module.conv1) is torch.nn.Conv2d


def test_maybe_replace_strided_conv_stride(state, blurpool_instance, dummy_logger):
    blurpool_instance.apply(Event.INIT, state, dummy_logger)
    assert isinstance(state.model.module, SimpleConvModel)

    assert type(state.model.module.conv3) is torch.nn.Conv2d  # stride = 1, should be no replacement


def test_maybe_replace_strided_conv_channels(state, blurpool_instance, dummy_logger):
    blurpool_instance.apply(Event.INIT, state, dummy_logger)
    assert isinstance(state.model.module, SimpleConvModel)

    assert type(state.model.module.conv2) is torch.nn.Conv2d  # channels < 16, should be no replacement


def test_blurconv_weights_preserved(state, blurpool_instance, dummy_logger):
    assert isinstance(state.model.module, SimpleConvModel)

    original_weights = state.model.module.conv1.weight.clone()
    blurpool_instance.apply(Event.INIT, state, dummy_logger)

    if isinstance(state.model.module.conv1, BlurConv2d):
        new_weights = state.model.module.conv1.conv.weight
    elif isinstance(state.model.module.conv1, torch.nn.Conv2d):
        new_weights = state.model.module.conv1.weight
    else:
        raise TypeError(f'Layer type {type(state.model.module.conv1)} not expected.')
    assert torch.allclose(original_weights, new_weights)


def test_blurpool(state, blurpool_instance, dummy_logger):
    blurpool_instance.apply(Event.INIT, state, dummy_logger)
    assert isinstance(state.model.module, SimpleConvModel)

    if blurpool_instance.hparams.replace_maxpools:
        assert type(state.model.module.pool1) is BlurMaxPool2d
    else:
        assert type(state.model.module.pool1) is torch.nn.MaxPool2d


def test_blurpool_wrong_event(state, blurpool_instance):
    assert blurpool_instance.match(Event.BATCH_START, state) == False


def test_blurpool_correct_event(state, blurpool_instance):
    assert blurpool_instance.match(Event.INIT, state) == True


def test_blurpool_algorithm_logging(state, blurpool_instance, dummy_logger):
    blurpool_instance.apply(Event.INIT, state, dummy_logger)

    dummy_logger.metric_fit.assert_called_once_with({
        'blurpool/num_blurpool_layers': 1 if blurpool_instance.hparams.replace_maxpools else 0,
        'blurpool/num_blurconv_layers': 1 if blurpool_instance.hparams.replace_convs else 0,
    })



# Copyright 2021 MosaicML. All Rights Reserved.

import pytest

from composer.utils.string_enum import StringEnum


def test_string_enum_invalid_name():
    with pytest.raises(ValueError):
        # names must be uppercase
        class TestStringEnum(StringEnum):
            names_must_be_uppercase = "names_must_be_uppercase"


def test_string_enum_invalid_value():
    with pytest.raises(ValueError):

        class TestStringEnum(StringEnum):
            VALUES_MUST_BE_LOWERCASE = "VALUES_MUST_BE_LOWERCASE"


def test_string_enum_comparision():

    class TestStringEnum(StringEnum):
        HELLO_WORLD = "hello_world"

    with pytest.warns(UserWarning):
        assert TestStringEnum.HELLO_WORLD == "hello_world"

    with pytest.warns(UserWarning):
        assert TestStringEnum.HELLO_WORLD == "HeLlO_WoRlD"


def test_missing():

    class TestStringEnum(StringEnum):
        HELLO_WORLD = "hello_world"

    real_val = TestStringEnum.HELLO_WORLD
    assert real_val == TestStringEnum(real_val)
    assert real_val == TestStringEnum("HeLlO_WoRlD")

    with pytest.raises(ValueError):
        TestStringEnum("unknown_name")

    with pytest.raises(TypeError):
        TestStringEnum(object())



# Copyright 2021 MosaicML. All Rights Reserved.

# type: ignore (pytorch code)
import warnings
from typing import List

from torch.optim.lr_scheduler import _LRScheduler


class WarmUpLR(_LRScheduler):
    """Decays the learning rate of each parameter group by either a small constant
    or linearly increasing small warmup factor until the number of epoch reaches a
    pre-defined milestone: :attr:`warmup_iters`.

    This scheduler is adapted from PyTorch but rewritten in a non-chainable form to
    accomodate :attr:`warmup_factor=0.0`. When :attr:`last_epoch=-1`, sets initial
    lr as lr.

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        warmup_factor (float): The number we multiply learning rate by in the first epoch.
            If the warming up method is constant, the multiplication factor of the
            learning rate stays the same in all epochs, but, in the linear case, it
            starts increasing in the following epochs. Default: ``1./3``.
        warmup_iters (int): The number of warming up steps. Default: ``5``.
        warmup_method (str): One of ``constant`` and ``linear``. In ``constant`` mode, the
            learning rate will be multiplied with a small constant until a milestone
            defined in :attr:``warmup_iters``. In the ``linear`` case, the multiplication factor
            starts with :attr:``warmup_factor`` in the first epoch then linearly increases to
            reach 1. in the epoch number :attr:``warmup_iters``. Default: ``linear``.
        last_epoch (int): The index of the last epoch. Can be used to restore the state of the
            learning rate schedule. Default: ``-1``.
        verbose (bool): If ``True``, prints a message to stdout for
            each update. Default: ``False``.
        interval (str): Frequency of ``step()`` calls, either ``step`` or ``epoch``. Default: ``step``.

    Example:
        >>> # Assuming optimizer uses lr = 0.05 for all groups
        >>> # lr = 0.025    if epoch == 0
        >>> # lr = 0.03125  if epoch == 1
        >>> # lr = 0.0375   if epoch == 2
        >>> # lr = 0.04375  if epoch == 3
        >>> # lr = 0.05    if epoch >= 4
        >>> scheduler = WarmUpLR(self.opt, warmup_factor=0.5, warmup_iters=4, warmup_method="linear")
        >>> for epoch in range(100):
        >>>     train(...)
        >>>     validate(...)
        >>>     scheduler.step()

        >>> # Assuming optimizer uses lr = 0.05 for all groups
        >>> # lr = 0.025    if epoch == 0
        >>> # lr = 0.025    if epoch == 1
        >>> # lr = 0.025    if epoch == 2
        >>> # lr = 0.025    if epoch == 3
        >>> # lr = 0.05    if epoch >= 4
        >>> scheduler = WarmUpLR(self.opt, warmup_factor=0.5, warmup_iters=4, warmup_method="constant")
        >>> for epoch in range(100):
        >>>     train(...)
        >>>     validate(...)
        >>>     scheduler.step()
    """

    def __init__(self,
                 optimizer,
                 warmup_factor=1.0 / 3,
                 warmup_iters=5,
                 warmup_method="linear",
                 last_epoch=-1,
                 verbose=False,
                 interval='step'):
        if warmup_method not in ("constant", "linear"):
            raise ValueError("Only 'constant' or 'linear' warmup_method accepted, but " "got {}".format(warmup_method))
        self.warmup_factor = warmup_factor
        self.warmup_iters = warmup_iters
        self.warmup_method = warmup_method
        self.interval = interval
        super(WarmUpLR, self).__init__(optimizer, last_epoch, verbose)

    def get_lr(self):
        """ Get the current learning rate for each parameter group.

        Returns:
            List of float: The current learning rate for each parameter group.
        """

        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, " "please use `get_last_lr()`.")

        if self.last_epoch == 0:
            return [group['lr'] * self.warmup_factor for group in self.optimizer.param_groups]

        if (self.last_epoch > self.warmup_iters or
            (self.warmup_method == "constant" and self.last_epoch != self.warmup_iters)):
            return [group['lr'] for group in self.optimizer.param_groups]

        if (self.warmup_method == "constant" and self.last_epoch == self.warmup_iters):
            return [group['lr'] * (1.0 / self.warmup_factor) for group in self.optimizer.param_groups]

        # to accomodate warmup_factor = 0, use non-chainable form of lr
        return [base_lr * self.lambda_lr() for base_lr in self.base_lrs]

    def lambda_lr(self) -> float:
        """ Compute the linear warmup ramp.

        Returns:
            float: Current warmup factor with which to scale the learning rate.
        """
        return (1 - self.warmup_factor) * (self.last_epoch / self.warmup_iters) + self.warmup_factor

    def _get_closed_form_lr(self) -> List[float]:
        """ Get the current learning rate for each parameter group.

        Returns:
            List of float: The current learning rate for each parameter group.
        """
        return [
            base_lr * (self.warmup_factor +
                       (1 - self.warmup_factor) * min(self.warmup_iters, self.last_epoch) / self.warmup_iters *
                       (self.warmup_method == "linear") + (self.last_epoch >= self.warmup_iters) *
                       (1 - self.warmup_factor) * (self.warmup_method == "constant")) for base_lr in self.base_lrs
        ]

    def scale_schedule(self, ssr: float) -> None:
        pass



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import collections.abc
import logging
import os
import subprocess
import sys
import time
import warnings
from abc import ABC, abstractmethod
from dataclasses import dataclass
from threading import Thread
from typing import Callable, Iterator, List, Optional, Sequence, TypeVar, cast

import torch
import torch.distributed
import torch.utils.data
import yahp as hp
from torch.nn.parallel import DistributedDataParallel
from torch.utils.data.distributed import DistributedSampler

from composer.core.state import State
from composer.core.types import Batch, DataLoader, Model, Tensor
from composer.datasets import DataloaderHparams, DataloaderSpec, WrappedDataLoader

logger = logging.getLogger(__name__)

TObj = TypeVar("TObj")


class DataloaderMultipleIterationWarning(Warning):
    pass


class DDPDataLoader(WrappedDataLoader):
    """
    DDPDataLoader wraps a dataloader and a distributed sampler to ensure that
    sampler.set_epoch() is called after each iteration (epoch) through the dataset
    See https://pytorch.org/docs/stable/data.html#torch.utils.data.distributed.DistributedSampler
    """

    def __init__(self, dataloader: DataLoader) -> None:
        super().__init__(dataloader)
        if not isinstance(self.dataloader.sampler, DistributedSampler):
            raise ValueError("When using the DDP data loader, the sampler must be a DistributedSampler")
        self._iterator: Optional[Iterator[Batch]] = None

    def __iter__(self) -> DDPDataLoader:
        if self._iterator is not None:
            warnings.warn(
                "The dataloader detected the start of a new iteration before the previous iteration finished. "
                "The dataloader is skipping ahead to the start of the next epoch. "
                "Multiple simultaneous iterations through the DDP dataloader prohibited, since "
                "it automatically tracks the current epoch.",
                category=DataloaderMultipleIterationWarning)
            assert isinstance(self.sampler, DistributedSampler)
            self.sampler.set_epoch(epoch=self.sampler.epoch + 1)
        self._iterator = iter(self.dataloader)
        return self

    def __next__(self) -> Batch:
        assert self._iterator is not None
        try:
            return next(self._iterator)
        except StopIteration:
            self._iterator = None
            assert isinstance(self.sampler, DistributedSampler)
            self.sampler.set_epoch(epoch=self.sampler.epoch + 1)
            raise


class DDP:

    def __init__(self,
                 *,
                 nproc_per_node: int,
                 store_hparams: StoreHparams,
                 node_rank: int,
                 num_nodes: int,
                 backend: str,
                 fork_rank_0: bool,
                 find_unused_parameters: bool = False):
        self.nproc_per_node = nproc_per_node
        self.world_size = num_nodes * nproc_per_node
        self.num_nodes = num_nodes
        self.node_rank = node_rank
        self.store_hparams = store_hparams
        self.last_return_code: Optional[int] = None
        self.backend = backend
        self.fork_rank_0 = fork_rank_0
        self.processes: List[subprocess.Popen[str]] = []
        self.find_unused_parameters = find_unused_parameters

        if backend == 'nccl':
            if not torch.cuda.is_available():
                raise ValueError('CUDA not available but gpu backend requested.')
            if torch.cuda.device_count() < nproc_per_node:
                raise ValueError(f'Requested {nproc_per_node} GPUs, but '\
                                 f'only {torch.cuda.device_count()} available.')
            if not torch.distributed.is_nccl_available():
                raise ValueError('Requested NCCL backend not available in torch.distributed')

    def barrier(self) -> None:
        if torch.distributed.is_available():
            torch.distributed.barrier()
        # If not on DDP, then do nothing

    def all_reduce(
        self,
        tensor: torch.Tensor,
        reduce_operation: str = "SUM",
    ) -> None:
        if torch.distributed.is_available():
            reduce_op = getattr(torch.distributed.ReduceOp, reduce_operation.upper())
            torch.distributed.all_reduce(tensor, op=reduce_op)
        else:
            # TODO implement NON-DDP versions of the reduce operations
            raise NotImplementedError("Non-DDP versions of reduce operations are not yet implemented")

    def all_gather(self, tensor: torch.Tensor) -> Sequence[Tensor]:
        """gather_to_rank_zero collects a tensor from each rank, and returns a sequence of tensors indexed by rank

        Args:
            tensor (torch.Tensor): tensor from each rank to be gathered

        Returns:
            Sequence[Tensor]: A sequence of tensors indexed by rank
        """
        if torch.distributed.is_available():
            obj_gather_list = [torch.zeros_like(tensor) for _ in range(self.world_size)]
            torch.distributed.all_gather(obj_gather_list, tensor)
            return obj_gather_list
        else:
            return [tensor]

    def all_gather_object(self, obj: TObj) -> List[TObj]:
        """gather_object_to_rank_zero collects a pickleable object from each rank, and returns a list of
        these objects indexed by rank

        Args:
            obj (TObj): Object to be gathered

        Returns:
            List[TObj]: A list of objects indexed by rank
        """
        if torch.distributed.is_available():
            obj_gather_list = [None for _ in range(self.world_size)]
            torch.distributed.all_gather_object(obj_gather_list, obj)
            # torch.distributed will replace the None's in obj_gather_list with the gathered objects on rank 0
            # or will just be None on non-rank-0
            return cast(List[TObj], obj_gather_list)
        else:
            return [obj]

    def launch(self, state: State, loop: Callable[[], None]):
        if os.environ.get("RANK") is None:
            os.environ["WORLD_SIZE"] = str(self.world_size)
            logger.info("Starting DDP on node_rank(%d) with world_size(%d)", self.node_rank, self.world_size)

            if torch.distributed.is_available():
                # Adapted from torch.distributed.launch

                # set PyTorch distributed related environmental variables

                current_env = os.environ.copy()
                # TODO omp num threads -- this parameter needs to be auto-tuned
                for local_rank in range(self.nproc_per_node):
                    # each process's rank
                    global_rank = self.nproc_per_node * self.node_rank + local_rank
                    current_env["RANK"] = str(global_rank)

                    if local_rank == 0 and not self.fork_rank_0:
                        os.environ["RANK"] = str(global_rank)
                    else:
                        logger.info("Launching process for global_rank(%d) on node_rank(%d)", global_rank,
                                    self.node_rank)
                        # spawn the processes
                        cmd = [
                            sys.executable,
                            "-u",
                            *sys.argv,
                        ]

                        if local_rank == 0:
                            # Attaching rank 0 to the main stdout/stderr so interactive terminal output will work without issue (e.g. tqdm)
                            process = subprocess.Popen(cmd, env=current_env, text=True)
                        else:
                            # Other processes, except in the case of an error, should not print anything
                            process = subprocess.Popen(
                                cmd,
                                env=current_env,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                text=True,
                            )
                        self.processes.append(process)
                if self.fork_rank_0:
                    self.monitor()
                    return
                else:
                    Thread(target=self.monitor, daemon=True).start()
            else:
                if self.world_size != 1:
                    raise ValueError("Must have world size == 1 when torch.distributed is not available")
                if self.node_rank != 0:
                    raise ValueError("Must have a node_rank == 0 when torch.distributed is not available")
                os.environ["RANK"] = "0"
        # We are now on the correct process
        global_rank = int(os.environ["RANK"])
        assert global_rank // self.world_size == self.node_rank
        assert os.environ["WORLD_SIZE"] == str(
            self.world_size
        ), f"os.environ['WORLD_SIZE']({os.environ['WORLD_SIZE']}) != self.world_size({self.world_size})"
        is_main = global_rank == 0
        if torch.distributed.is_available():
            logger.info("Initializing ddp: GLOBAL_RANK: %s, WORLD_SIZE: %s", global_rank, self.world_size)
            store = self.store_hparams.initialize_object(is_main, state.world_size)
            torch.distributed.init_process_group(self.backend,
                                                 rank=global_rank,
                                                 world_size=self.world_size,
                                                 store=store)
            assert torch.distributed.is_initialized()
            assert state.is_rank_set, "state.is_rank_set should be set after torch.distributed is initialized"
            assert state.local_rank == global_rank % self.nproc_per_node, "state.local_rank is incorrect"
            assert state.nproc_per_node == self.nproc_per_node, "state.nproc_per_node is incorrect"
            assert state.global_rank == torch.distributed.get_rank(
            ), "state.global_rank != torch.distributed.get_rank()"
            logger.info("All DDP processes registered. world_size=%s.", self.world_size)
            logger.info("Starting process with global_rank=%s", global_rank)
        try:
            loop()
        finally:
            self.cleanup()

    def prepare_module(self, module: Model) -> Model:
        if torch.distributed.is_available():
            if any((p.requires_grad for p in module.parameters())):
                ddp_model = DistributedDataParallel(module, find_unused_parameters=self.find_unused_parameters)
                return cast(Model, ddp_model)
            return module
        else:
            return module

    def create_dataloader(self, batch_size: int, dataloader_hparams: DataloaderHparams,
                          dataloader_spec: DataloaderSpec) -> DataLoader:
        if torch.distributed.is_available():
            sampler = torch.utils.data.DistributedSampler[int](dataloader_spec.dataset,
                                                               drop_last=dataloader_spec.drop_last,
                                                               shuffle=dataloader_spec.shuffle)
        else:
            assert isinstance(dataloader_spec.dataset, collections.abc.Sized)
            sampler = torch.utils.data.RandomSampler(dataloader_spec.dataset, generator=dataloader_spec.generator)
        dataloader = dataloader_hparams.initialize_object(batch_size, sampler, dataloader_spec)
        if torch.distributed.is_available():
            dataloader = DDPDataLoader(dataloader)
        return dataloader

    def monitor(self) -> None:
        # Monitor checks whether any subprocesses have died unexpectedly
        alive_processes = set(self.processes)
        while len(alive_processes) > 0:
            finished_processes: List[subprocess.Popen[str]] = []
            for process in alive_processes:
                if process.poll() is None:
                    # the process is still running
                    continue
                else:
                    if process.returncode != 0:
                        if process.stdout is None:
                            output = ""
                        else:
                            output = process.stdout.read()

                        if process.stderr is None:
                            stderr = ""
                        else:
                            stderr = process.stderr.read()
                        exc = subprocess.CalledProcessError(
                            process.returncode,
                            cmd=process.args,
                            output=output,
                            stderr=stderr,
                        )
                        if self.fork_rank_0:
                            raise exc
                        else:
                            logger.exception("Error in subprocess", exc_info=exc)
                            sys.exit(1)
                    else:
                        # exited cleanly
                        finished_processes.append(process)
            alive_processes = set(alive_processes) - set(finished_processes)
            time.sleep(1)

    def cleanup(self) -> None:
        for process in self.processes:
            logger.info("Killing subprocess %s", process.pid)
            try:
                process.kill()
            except Exception:
                pass
        if torch.distributed.is_initialized():
            torch.distributed.destroy_process_group()


@dataclass
class StoreHparams(hp.Hparams, ABC):

    @abstractmethod
    def initialize_object(self, is_main: bool, world_size: int) -> torch.distributed.Store:
        pass


@dataclass
class TCPStoreHparams(StoreHparams):
    host_name: str = hp.optional(doc="Rank 0 address", default="127.0.0.1")
    port: int = hp.optional(doc="Rank 0 port", default=43297)

    def initialize_object(self, is_main: bool, world_size: int) -> torch.distributed.Store:
        return torch.distributed.TCPStore(self.host_name, self.port, world_size, is_main)


@dataclass
class FileStoreHparams(StoreHparams):
    file_name: str = hp.required(doc="Path to store file")

    def initialize_object(self, is_main: bool, world_size: int) -> torch.distributed.Store:
        return torch.distributed.FileStore(self.file_name, world_size)


@dataclass
class DDPHparams(hp.Hparams):
    hparams_registry = {
        "store": {
            "tcp": TCPStoreHparams,
            "file": FileStoreHparams,
        }
    }

    store: StoreHparams = hp.optional(doc="Store", default_factory=TCPStoreHparams)
    node_rank: int = hp.optional(doc="Node ID for multi-node training", default=0)
    num_nodes: int = hp.optional(doc="Number of nodes used for training", default=1)
    fork_rank_0: bool = hp.optional(
        doc="Whether to fork the local rank 0 process, or use the existing process for rank 0 training.",
        default=True,
    )

    def initialize_object(self, nproc_per_node: int, backend: str, find_unused_parameters: bool) -> DDP:
        return DDP(
            backend=backend,
            nproc_per_node=nproc_per_node,
            store_hparams=self.store,
            node_rank=self.node_rank,
            num_nodes=self.num_nodes,
            fork_rank_0=self.fork_rank_0,
            find_unused_parameters=find_unused_parameters,
        )



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.core import types as types
from composer.core.algorithm import Algorithm as Algorithm
from composer.core.callback import Callback as Callback
from composer.core.engine import Engine as Engine
from composer.core.engine import Trace as Trace
from composer.core.event import Event as Event
from composer.core.logging import Logger as Logger
from composer.core.state import State as State



# Copyright 2021 MosaicML. All Rights Reserved.

import setuptools
from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

install_requires = [
    "pyyaml>=5.4.1",
    "tqdm>=4.62.3",
    "torchmetrics",
    "torch_optimizer",
    "torchvision>=0.9.0",
    "torch>=1.8.1",
    "argparse>=1.4.0",
    "yahp>=0.0.10",
]
extra_deps = {}

extra_deps['base'] = []

extra_deps['dev'] = [
    'junitparser>=2.1.1',
    'coverage[toml]>=5.5',
    'pytest>=6.2.0',
    'yapf>=0.13.0',
    'isort>=5.9.3',
    'yamllint>=1.26.2',
    'pytest-timeout>=1.4.2',
    'recommonmark>=0.7.1',
    'sphinx>=4.2.0',
    'sphinx_copybutton>=0.4.0',
    'sphinx_markdown_tables>=0.0.15',
    'sphinxcontrib.katex>=0.8.6',
    'sphinxext.opengraph>=0.4.2',
    'sphinx_rtd_theme>=1.0.0',
    'myst-parser>=0.15.2',
]
extra_deps['wandb'] = ['wandb>=0.12.2']

extra_deps['nlp'] = [
    'transformers',
    'datasets',
]

extra_deps['unet'] = [
    'medpy==0.4.0',
    'monai',
    'scikit-learn',
    'torch_optimizer>=0.1.0',
]

extra_deps['all'] = set(dep for deps in extra_deps.values() for dep in deps)

# TODO fix deferred loading in composer.functional
install_requires.extend(extra_deps['unet'])
install_requires.extend(extra_deps['nlp'])

setup(
    name="mosaicml",
    version="0.2.3",
    author="MosaicML",
    author_email="team@mosaicml.com",
    description="composing methods for ML training efficiency",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mosaicml/composer",
    include_package_data=True,
    package_data={
        "": ["*.yaml", "*.yml"],
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    install_requires=install_requires,
    entry_points={
        'console_scripts': ['composer = examples.run_mosaic_trainer:main',],
    },
    extras_require=extra_deps,
    dependency_links=['https://developer.download.nvidia.com/compute/redist'],
    python_requires='>=3.8',
    ext_package="composer",
)

# only visible if user installs with verbose -v flag
print("*" * 20)
print("\nNOTE: For best performance, we recommend installing Pillow-SIMD "
      "\nfor accelerated image processing operations. To install:"
      "\n\n\t pip uninstall pillow && pip install pillow-simd\n")
print("*" * 20)



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import sys
from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence

import yaml
from tqdm import tqdm

from composer.core.logging import LogLevel, RankZeroLoggerBackend, TLogData, TLogDataValue, format_log_data_value
from composer.core.state import State
from composer.core.types import StateDict

if TYPE_CHECKING:
    from composer.core.logging import Logger


@dataclass
class TQDMLoggerInstanceState:
    total: int
    epoch: int
    val: bool
    n: int
    keys_to_log: Sequence[str]
    epoch_metrics: Dict[str, TLogDataValue] = field(default_factory=dict)


class TQDMLoggerInstance:

    def __init__(self,
                 total: int,
                 epoch: int,
                 val: bool,
                 keys_to_log: Sequence[str],
                 n: int = 0,
                 epoch_metrics: Optional[Dict[str, TLogDataValue]] = None) -> None:
        self.state = TQDMLoggerInstanceState(total=total,
                                             epoch=epoch,
                                             val=val,
                                             n=n,
                                             keys_to_log=keys_to_log,
                                             epoch_metrics=(epoch_metrics or {}))
        desc = f'Epoch {epoch + 1}{" (val)" if val else ""}'
        position = 1 if val else 0
        self.pbar = tqdm(total=total,
                         desc=desc,
                         position=position,
                         initial=n,
                         bar_format="{l_bar}{bar:10}{r_bar}{bar:-10b}")
        self.pbar.set_postfix(epoch_metrics)

    def log_metric(self, data: TLogData):
        formatted_data = {k: format_log_data_value(v) for (k, v) in data.items() if k in self.state.keys_to_log}
        self.state.epoch_metrics.update(formatted_data)
        self.pbar.set_postfix(self.state.epoch_metrics)

    def update(self):
        self.pbar.update()
        self.state.n = self.pbar.n

    def close(self):
        self.pbar.close()

    def state_dict(self) -> StateDict:
        return asdict(self.state)


class TQDMLoggerBackend(RankZeroLoggerBackend):

    def __init__(self, config: Optional[Dict[str, Any]] = None) -> None:
        super().__init__()
        self.pbar_train: Optional[TQDMLoggerInstance] = None
        self.pbar_val: Optional[TQDMLoggerInstance] = None
        self.is_validating = False
        self.config = config

    def _will_log(self, state: State, log_level: LogLevel) -> bool:
        return log_level <= LogLevel.BATCH

    def _log_metric(self, epoch: int, step: int, log_level: LogLevel, data: TLogData) -> None:
        pbar = self.pbar_val if self.is_validating else self.pbar_train
        if pbar is None:
            # Logging outside an epoch
            return
        pbar.log_metric(data)

    def _training_start(self, state: State, logger: Logger) -> None:
        if self.config is not None:
            print("Config")
            print("-" * 30)
            yaml.safe_dump(self.config, stream=sys.stdout)
            print("-" * 30)
            print()

    def epoch_start(self, state: State, logger: Logger) -> None:
        assert self.pbar_train is None
        self.pbar_train = TQDMLoggerInstance(total=state.steps_per_epoch,
                                             epoch=state.epoch,
                                             val=False,
                                             keys_to_log=["loss/train"])

    def after_backward(self, state: State, logger: Logger) -> None:
        assert self.pbar_train is not None
        self.pbar_train.update()

    def epoch_end(self, state: State, logger: Logger) -> None:
        assert self.pbar_train is not None
        self.pbar_train.close()
        self.pbar_train = None

    def eval_start(self, state: State, logger: Logger) -> None:
        assert self.pbar_val is None
        assert state.eval_dataloader is not None
        self.pbar_val = TQDMLoggerInstance(total=len(state.eval_dataloader),
                                           epoch=state.epoch,
                                           val=True,
                                           keys_to_log=["accuracy/val"])
        self.is_validating = True

    def eval_after_forward(self, state: State, logger: Logger) -> None:
        assert self.pbar_val is not None
        self.pbar_val.update()

    def eval_end(self, state: State, logger: Logger) -> None:
        assert self.pbar_val is not None
        self.pbar_val.close()
        self.pbar_val = None
        self.is_validating = False

    def state_dict(self) -> StateDict:
        state = {"is_validating": self.is_validating}
        if self.pbar_train:
            state["pbar_train"] = self.pbar_train.state_dict()
        if self.pbar_val:
            state["pbar_val"] = self.pbar_val.state_dict()
        return state

    def load_state_dict(self, state: StateDict) -> None:
        self.is_validating = state["is_validating"]
        if "pbar_train" in state:
            self.pbar_train = TQDMLoggerInstance(**state["pbar_train"])
        if "pbar_val" in state:
            self.pbar_val = TQDMLoggerInstance(**state["pbar"])



# Copyright 2021 MosaicML. All Rights Reserved.

from unittest.mock import Mock

import pytest
import torch
import torch.nn.functional as F

from composer.algorithms import label_smoothing
from composer.algorithms.label_smoothing import LabelSmoothingHparams
from composer.core.types import Event
from composer.models import loss
from composer.trainer.trainer_hparams import TrainerHparams
from tests.utils.trainer_fit import train_model


def _generate_tensors_classification(batch_size: int, num_classes: int):
    """
    Helper functions to generate input, target pairs
    for image classification (1d indicies)
    """
    N = batch_size
    C = num_classes

    target_indicies = torch.randint(0, C, [N])
    target_onehot = F.one_hot(target_indicies, num_classes=C)
    input = F.softmax(torch.randn((N, C)), dim=1)

    return (input, target_indicies, target_onehot)


def _generate_tensors_segmentation(batch_size: int, num_classes: int, H: int, W: int):
    """
    Helper functions to generate input, target pairs
    for image segmentation (2d indicies)
    """
    N = batch_size
    C = num_classes

    target_indicies = torch.randint(0, C, (N, H, W))
    target_onehot = F.one_hot(target_indicies, num_classes=C)  # NHWC?
    input = F.softmax(torch.randn((N, C, H, W)), dim=1)

    return (input, target_indicies, target_onehot)


def xfail(val):
    """
    shorthand to mark xfail parameters
    """
    return pytest.param(val, marks=pytest.mark.xfail)


def generate_tensors():
    return [
        # binary classification
        _generate_tensors_classification(batch_size=64, num_classes=2),
        # classification
        _generate_tensors_classification(batch_size=64, num_classes=10),
        # segmentation
        xfail(_generate_tensors_segmentation(batch_size=64, num_classes=2, H=5, W=5)),
        xfail(_generate_tensors_segmentation(batch_size=64, num_classes=10, H=5, W=5))
    ]


@pytest.mark.parametrize('tensors', generate_tensors())
class TestSoftCrossEntropy:

    def test_infer_target_type(self, tensors):
        (input, target_indicies, target_onehot) = tensors
        assert loss._infer_target_type(input, target_indicies) == 'indicies'
        assert loss._infer_target_type(input, target_onehot) == 'one_hot'

    @pytest.mark.parametrize('reduction', ['mean', 'sum'])
    @pytest.mark.parametrize('use_weights', [xfail(True), False])
    def test_soft_cross_entropy(self, tensors, use_weights, reduction):
        (input, target_indicies, target_onehot) = tensors
        if use_weights:
            num_classes = target_onehot.shape[-1]
            weights = torch.randn(size=[num_classes])
        else:
            weights = None

        loss_args = dict(weight=weights, reduction=reduction)

        loss_indicies = loss.soft_cross_entropy(input, target_indicies, **loss_args)
        loss_onehot = loss.soft_cross_entropy(input, target_onehot, **loss_args)
        loss_reference = F.cross_entropy(input, target_indicies, **loss_args)

        torch.testing.assert_allclose(loss_indicies, loss_onehot)
        torch.testing.assert_allclose(loss_indicies, loss_reference)


@pytest.mark.parametrize('alpha', [0, 0.1, 0.5, 0.9, 1.0])
@pytest.mark.parametrize('tensors', generate_tensors())
class TestLabelSmoothing:

    @staticmethod
    def reference_smooth_labels(targets, alpha):
        num_classes = targets.shape[-1]
        return targets * (1 - alpha) + alpha / num_classes

    def test_label_smoothing(self, tensors, alpha):
        (input, target_indicies, target_onehot) = tensors

        labels_onehot = label_smoothing.smooth_labels(input, target_onehot, alpha)
        labels_indicies = label_smoothing.smooth_labels(input, target_indicies, alpha)
        labels_ref = self.reference_smooth_labels(target_onehot, alpha)

        torch.testing.assert_allclose(labels_onehot, labels_ref)
        torch.testing.assert_allclose(labels_indicies, labels_ref)

    @pytest.mark.parametrize('target_type', ['onehot', 'indicies'])
    def test_label_smoothing_algorithm(self, tensors, alpha, target_type, dummy_logger, dummy_state):
        (outputs, target_indicies, target_onehot) = tensors

        target = target_indicies if target_type == 'indicies' else target_onehot

        algorithm = LabelSmoothingHparams(alpha=alpha).initialize_object()
        state = dummy_state
        state.batch = (torch.Tensor(), target)
        state.outputs = outputs

        # BEFORE_LOSS should smooth the labels
        algorithm.apply(Event.BEFORE_LOSS, state, dummy_logger)
        smoothed_reference = self.reference_smooth_labels(target_onehot, alpha)

        _, labels = state.batch
        torch.testing.assert_allclose(labels, smoothed_reference)

        # AFTER_LOSS should restore the original targets
        algorithm.apply(Event.AFTER_LOSS, state, dummy_logger)

        _, labels = state.batch
        torch.testing.assert_allclose(labels, target)


def test_label_smoothing_match():
    algorithm = LabelSmoothingHparams(alpha=0.1).initialize_object()
    assert algorithm.match(Event.BEFORE_LOSS, Mock())
    assert algorithm.match(Event.AFTER_LOSS, Mock())


@pytest.mark.run_long
@pytest.mark.timeout(90)
def test_label_smoothing_trains(mosaic_trainer_hparams: TrainerHparams):
    mosaic_trainer_hparams.algorithms = [LabelSmoothingHparams(alpha=0.1)]
    train_model(mosaic_trainer_hparams)



# Copyright 2021 MosaicML. All Rights Reserved.

import logging
from dataclasses import asdict, dataclass
from typing import Optional

import numpy as np
import torch
import yahp as hp
from torch.nn import functional as F

from composer.algorithms import AlgorithmHparams
from composer.core.types import Algorithm, Event, Logger, State, Tensor

log = logging.getLogger(__name__)


def gen_interpolation_lambda(alpha: float) -> float:
    # First check if alpha is positive.
    assert alpha >= 0
    # Draw the interpolation parameter from a beta distribution.
    # Check here is needed because beta distribution requires alpha > 0
    # but alpha = 0 is fine for mixup.
    if alpha == 0:
        interpolation_lambda = 0
    else:
        interpolation_lambda = np.random.beta(alpha, alpha)
    # for symmetric beta distribution, can always use 0 <= lambda <= .5;
    # this way the "main" label is always the original one, which keeps
    # the training accuracy meaningful
    return max(interpolation_lambda, 1. - interpolation_lambda)


def mixup_batch(x: Tensor,
                y: Tensor,
                interpolation_lambda: float,
                n_classes: int,
                indices: Optional[torch.Tensor] = None):
    """Implements mixup on a single batch of data.

    This constructs a new batch of data given an original batch.
    This is done through the convex combination of x with a randomly
    permuted copy of x. The interploation parameter lambda should be chosen from
    a beta distribution with parameter alpha. Note that the same lambda is
    used for all examples within the batch.

    Both the original and shuffled labels are returned. This is done because
    for many loss functions (such as cross entropy) the targets are given as
    indices, so interpolation must be handled separately.

    Args:
        x: Input tensor of shape (B, d1, d2, ..., dn), B is batch size, d1-dn
            are feature dimensions.
        y: Target tensor of shape (B, f1, f2, ..., fm), B is batch size, f1-fn
            are possible target dimensions.
        interpolation_lambda: Amount of interpolation based on alpha.
        n_classes: Total number of classes.
        indices: Tensor of shape (B). Permutation of the batch indices. Used
            for permuting without randomness.

    Returns:
        x_mix: Batch of inputs after mixup has been applied.
        y_mix: Labels after mixup has been applied.
    """
    # Create shuffled versions of x and y in preparation for interpolation
    # Use given indices if there are any.
    if indices is None:
        shuffled_idx = torch.randperm(x.shape[0])
    else:
        shuffled_idx = indices
    x_shuffled = x[shuffled_idx]
    y_shuffled = y[shuffled_idx]
    # Interpolate between the inputs
    x_mix = (1 - interpolation_lambda) * x + interpolation_lambda * x_shuffled

    y_onehot = F.one_hot(y, num_classes=n_classes)
    y_shuffled_onehot = F.one_hot(y_shuffled, num_classes=n_classes)
    y_mix = ((1. - interpolation_lambda) * y_onehot + interpolation_lambda * y_shuffled_onehot)

    return x_mix, y_mix, shuffled_idx


@dataclass
class MixUpHparams(AlgorithmHparams):

    alpha: float = hp.required('Strength of interpolation, should be >= 0. No interpolation if alpha=0.',
                               template_default=0.2)

    def initialize_object(self) -> "MixUp":
        return MixUp(**asdict(self))


class MixUp(Algorithm):
    """Applies MixUp algorithm by modifying the images and labels during Event.AFTER_DATALOADER."""

    def __init__(self, alpha: float):
        self.hparams = MixUpHparams(alpha=alpha)

    def match(self, event: Event, state: State) -> bool:
        return event in (Event.AFTER_DATALOADER, Event.INIT)

    @property
    def interpolation_lambda(self) -> float:
        return self._interpolation_lambda

    @interpolation_lambda.setter
    def interpolation_lambda(self, new_int_lamb) -> None:
        self._interpolation_lambda = new_int_lamb

    @property
    def indices(self) -> Tensor:
        return self._indices

    @indices.setter
    def indices(self, new_indices: Tensor) -> None:
        self._indices = new_indices

    def apply(self, event: Event, state: State, logger: Logger) -> None:
        if event == Event.INIT:
            # TODO(issue #249): Fix
            self.num_classes: int = state.model.num_classes  # type: ignore
            return

        input, target = state.batch_pair
        assert isinstance(input, Tensor) and isinstance(target, Tensor), \
            "Multiple tensors for inputs or targets not supported yet."
        alpha = self.hparams.alpha

        self.interpolation_lambda = gen_interpolation_lambda(alpha)

        new_input, new_target, self.indices = mixup_batch(
            x=input,
            y=target,
            interpolation_lambda=self.interpolation_lambda,
            n_classes=self.num_classes,
        )

        state.batch = (new_input, new_target)



# Copyright 2021 MosaicML. All Rights Reserved.

import argparse
import logging
import math

import numpy as np
import yaml


def teraflops_for_accelerator(accel):
    """
    Stores the number of TFLOPs available to a few accelerators, including driver handicaps.

    Args:
        accel (str): A string descriptor of which accelerator to use. Must be either "3090" or "V100".

    Returns:
        accel_flops (int): an integer of how many TFLOPs are in the accelerator.
    """
    accel_flops = {"3090": 71, "V100": 125}
    return accel_flops[accel]


def parse_args():
    """
    ArgParse parser.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--hours",
                        type=float,
                        help="Number of hours to train for in order to set a total FLOPs budget.")
    parser.add_argument("--accelerator_for_budget",
                        choices=['3090', 'V100'],
                        type=str,
                        default="3090",
                        help="Hardware accelerator used to calculate TFLOPs.")
    parser.add_argument(
        "--train_sequence_length",
        type=int,
        default=1024,
        help="Fixed sequence length to train on, used for calculating how many batches and serial steps to trian for.")
    parser.add_argument("--warmup_ratio",
                        type=float,
                        default=0.1,
                        help="What % of total training batches to warm up the learning rate for.")
    parser.add_argument("--per_device_batch_size",
                        type=int,
                        default=7,
                        help="Number of training examples to fit per-device.")
    parser.add_argument("--num_devices", type=int, default=8, help="Number of training accelerators to use.")
    parser.add_argument("--output_file", type=str, help="Path to output a YAML file detailing the configuration.")
    parser.add_argument(
        "--no_grad_accum",
        default=False,
        action="store_true",
        help="Whether to ignore num_serial_steps predictions and train at smaller optimization batch sizes.")
    parser.add_argument("--validation_freq",
                        default=0.05,
                        type=float,
                        help="Percentage of training steps to run validation for.")
    parser.add_argument("--ssr", help="Scale Schedule Ratio", type=float, default=1.0)
    parser.add_argument("--num_validation_tokens", help="Number of validation tokens", type=int, default=35262464)
    return parser.parse_args()


args = parse_args()
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

model = {
    "activation_function": "gelu_new",
    "architectures": ["GPT2LMHeadModel"],
    "attn_pdrop": 0.1,
    "bos_token_id": 50256,
    "embd_pdrop": 0.1,
    "eos_token_id": 50256,
    "initializer_range": 0.02,
    "layer_norm_epsilon": 1e-05,
    "model_type": "gpt2",
    "n_ctx": 1024,
    "n_embd": 768,
    "n_head": 12,
    "n_inner": None,
    "n_layer": 12,
    "n_positions": 1024,
    "resid_pdrop": 0.1,
    "scale_attn_weights": True,
    "summary_activation": None,
    "summary_first_dropout": 0.1,
    "summary_proj_to_labels": True,
    "summary_type": "cls_index",
    "summary_use_proj": True,
    "task_specific_params": {
        "text-generation": {
            "do_sample": True,
            "max_length": 50
        }
    },
    "transformers_version": "4.11.0.dev0",
    "use_cache": True,
    "vocab_size": 50257
}

template_yaml = {
    'train_dataset': {
        'lm': {
            'split': 'train',
            'datadir': ['/datasets/openwebtext_saved'],
            'tokenizer_name': 'gpt2',
            'seed': 17,
            'shuffle': False,
            'drop_last': False
        }
    },
    'val_dataset': {
        'lm': {
            'split': 'validation',
            'datadir': ['/datasets/openwebtext_saved'],
            'tokenizer_name': 'gpt2',
            'seed': 17,
            'shuffle': False,
            'drop_last': False
        }
    },
    'model': {
        'gpt2': {
            'use_pretrained': False,
            "tokenizer_name": "gpt2",
            'model_config': model
        }
    },
    'optimizer': {
        'adamw': {
            'lr': 0.0003,
            'betas': [0.9, 0.999],
            'eps': 1e-06,
            'weight_decay': 0.0
        }
    },
    'schedulers': [{
        'warmup': {
            'warmup_method': 'linear',
            'warmup_factor': 0,
            'interval': 'step',
            'warmup_iters': 1
        }
    }, {
        'cosine_decay': {
            'T_max': 2,
            'interval': 'step',
            'eta_min': 1.0e-5,
            'verbose': False
        }
    }],
    'loggers': [
        {
            'file': {
                'log_level': 'BATCH',
                'filename': 'stdout',
                'buffer_size': 1,
                'flush_every_n_batches': 100,
                'every_n_epochs': 1,
                'every_n_batches': 100
            }
        },
        {
            'wandb': {
                "project": "gpt2",
                "name": f"gpt2-{args.hours}-ga-{not args.no_grad_accum}",
                'extra_init_params': {}
            }
        },
    ],
    'max_epochs': 1,
    'total_batch_size': 8,
    'eval_batch_size': 8,
    'seed': 17,
    'accelerator': {
        'gpu': {
            'n_gpus': 1,
            'prefetch_in_cuda_stream': False,
        }
    },
    'dataloader': {
        'pin_memory': True,
        'persistent_workers': True,
        'num_workers': 8,
        'timeout': 0,
        'prefetch_factor': 2
    },
    'grad_accum': 1,
    'precision': 'amp',
    'grad_clip_norm': 1.0,
}


def generate_architecture(args, model):
    """
    Given the desired training budget and a template model, configure the model archtiecture according to
    "Scaling Laws for Neural Language Models" by Kaplan et al.

    Args:
        args (argparse.Namespace): the Namespace object holding the parsed arguments.
        model (Mapping): a dictionary of a base model to configure.
    """
    accelerator_hours = args.hours
    practical_efficiency = 1.0 / 4.0
    hours_to_day = 1.0 / 24.0
    teraflops = teraflops_for_accelerator(args.accelerator_for_budget)
    teraflops_to_petaflops = 1.0 / 10**3

    desired_petaflop_days = teraflops * practical_efficiency * \
        hours_to_day * teraflops_to_petaflops * accelerator_hours
    petaflop_day_const = 3.1 * 10**8
    compute_exp = 0.050
    predicted_loss = (petaflop_day_const / desired_petaflop_days)**compute_exp

    dataset_exp = 0.27
    dataset_const = 2 * 10**10
    min_dataset_size = dataset_const * desired_petaflop_days**dataset_exp
    min_dataset_size = round(min_dataset_size)

    model_const = 1.3 * 10**9
    model_exp = 0.73
    min_model_size = model_const * desired_petaflop_days**model_exp
    min_model_size = round(min_model_size)

    serial_const = 5.4 * 10**3
    serial_exp = 0.03
    num_serial_steps = serial_const * desired_petaflop_days**serial_exp
    num_serial_steps = round(num_serial_steps)

    expected_lr = 0.003239 - (0.0001395 * np.log(min_model_size))

    logger.info(f"For a compute budget of {accelerator_hours} GPU hours:")
    logger.info("----------------- GENERAL PARAMETERS -----------------")
    logger.info(f"Predicted loss: {predicted_loss}")
    logger.info(f"Predicted perplexity: {np.exp(predicted_loss)}")
    logger.info(f"Predicted minimum dataset size: {min_dataset_size:,}")
    logger.info(f"Predicted minimum model size: {min_model_size:,}")
    logger.info(f"Predicted minimum serial steps: {num_serial_steps:,}")
    logger.info(f"Predicted learning rate: {expected_lr:e}")
    logger.info("\n")

    # See Eqn. 2.1 from https://arxiv.org/pdf/2001.08361.pdf for derivation of these equations
    d_model = ((min_model_size) * 100.0) / 12.0
    d_model = d_model**(1.0 / 3.0)
    n_layers = d_model / 100.0

    n_layers = math.ceil(n_layers)

    feedforward_ratio = 4
    d_ff = feedforward_ratio * d_model

    attn_head_ratio = 50
    n_head = d_model / attn_head_ratio
    n_head = math.ceil(n_head)

    # make sure d_model is a multiple of n_head and gpu tile size
    gpu_tile_size = 8
    d_model = (n_head * gpu_tile_size) * int(round(d_model / (n_head * gpu_tile_size)))
    d_ff = gpu_tile_size * round(d_ff / gpu_tile_size)

    logger.info("----------------- MODEL ARCHITECTURE DETAILS -----------------")
    logger.info(f"Number of layers: {n_layers}")
    logger.info(f"Hidden dimension: {d_model}")
    logger.info(f"Number of attention heads: {n_head}")
    logger.info(f"Feedforward dimension: {d_ff}")

    model['n_embd'] = d_model
    model['n_head'] = n_head
    model['n_layer'] = n_layers
    model['n_inner'] = d_ff

    scaling_law_predictions = {}
    scaling_law_predictions['pred_loss'] = predicted_loss
    scaling_law_predictions['pred_ppl'] = np.exp(predicted_loss)
    scaling_law_predictions['pred_min_dataset_size'] = min_dataset_size
    scaling_law_predictions["pred_min_model_size"] = min_model_size
    scaling_law_predictions['pred_min_serial_steps'] = num_serial_steps
    scaling_law_predictions['pred_lr'] = float(expected_lr)
    return model, scaling_law_predictions


def configure_mosaic_yaml(model, scaling_law_predictions):
    template_yaml['optimizer']['adamw']['lr'] = scaling_law_predictions['pred_lr']

    logger.info("----------------- OPTIMIZATION INFORMATION -----------------")
    logger.info(f"Number of tokens: {scaling_law_predictions['pred_min_dataset_size']:.4e}")
    num_training_tokens = math.ceil(scaling_law_predictions['pred_min_dataset_size'])

    curr_num_batches = num_training_tokens // args.train_sequence_length
    min_serial_steps = scaling_law_predictions['pred_min_serial_steps']
    batch_size = args.per_device_batch_size * args.num_devices
    curr_serial_steps = math.ceil(curr_num_batches / batch_size)

    # we ignore the grad accum paramters to make Mosaic Trainer easier to work with
    if args.no_grad_accum:
        lr_scaling_factor = math.floor(curr_serial_steps / min_serial_steps)
        template_yaml['optimizer']['adamw']['lr'] = template_yaml['optimizer']['adamw']['lr'] / lr_scaling_factor
        curr_grad_accum = 1
    else:
        curr_grad_accum = int(math.floor(curr_serial_steps / min_serial_steps))

    batch_size = args.per_device_batch_size * args.num_devices * curr_grad_accum
    orig_serial_steps = math.ceil(curr_num_batches / batch_size)
    logger.info(f"Applying Scale Schedule Ratio = {args.ssr}")
    final_serial_steps = math.ceil(orig_serial_steps * args.ssr)

    total_tokens_trained = args.train_sequence_length * batch_size * final_serial_steps

    # permit a 1% difference in number of training tokens in order to ensure we don't drop batches
    assert abs((((total_tokens_trained) / args.ssr) - num_training_tokens) / num_training_tokens) < 1e-2
    num_training_tokens = total_tokens_trained
    logger.info(f"New number of tokens: {num_training_tokens:.4e}")

    template_yaml['train_dataset']['lm']['num_tokens'] = num_training_tokens
    template_yaml['val_dataset']['lm']['num_tokens'] = args.num_validation_tokens

    logger.info(f"Total batch size: {batch_size:,}")
    logger.info(f"Total grad accum: {curr_grad_accum:,}")
    logger.info(f"Minumum possible serial optimization steps before SSR: {min_serial_steps:,}")
    logger.info(f"Minumum possible serial optimization steps after SSR: {math.ceil(args.ssr * min_serial_steps):,}")
    logger.info(f"Current serial optimization steps: {final_serial_steps:,}")
    template_yaml['total_batch_size'] = batch_size
    assert math.floor(batch_size / curr_grad_accum) == (batch_size / curr_grad_accum)
    template_yaml['eval_batch_size'] = math.floor(batch_size / curr_grad_accum)
    template_yaml['grad_accum'] = curr_grad_accum
    template_yaml['accelerator']['gpu']['n_gpus'] = args.num_devices
    warmup_steps = round(orig_serial_steps * args.warmup_ratio)
    decay_steps = final_serial_steps - warmup_steps
    template_yaml['schedulers'][0]['warmup']['warmup_iters'] = f"{warmup_steps}ba"
    template_yaml['schedulers'][1]['cosine_decay']['T_max'] = f"{decay_steps}ba"
    template_yaml['model']['gpt2']['model_config'] = model

    validation_freq = math.floor(final_serial_steps * args.validation_freq)

    return template_yaml


if __name__ == "__main__":
    model, scaling_law_predictions = generate_architecture(args, model)
    template_yaml = configure_mosaic_yaml(model, scaling_law_predictions)

    with open(args.output_file, "w+") as f:
        yaml.dump(template_yaml, f, sort_keys=False)



# Copyright 2021 MosaicML. All Rights Reserved.

"""
Example usage and definition of hparams
"""
from __future__ import annotations

import os
from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Optional

import yahp as hp

import composer
import composer.datasets as datasets
from composer.algorithms import AlgorithmHparams, get_algorithm_registry
from composer.callbacks import (BenchmarkerHparams, CallbackHparams, GradMonitorHparams, LRMonitorHparams,
                                SpeedMonitorHparams, TorchProfilerHparams)
from composer.core.types import Precision
from composer.datasets import DataloaderHparams
from composer.loggers import BaseLoggerBackendHparams, logger_registry
from composer.models import (CIFARResNetHparams, EfficientNetB0Hparams, GPT2Hparams, MnistClassifierHparams,
                             ModelHparams, ResNet18Hparams, ResNet50Hparams, ResNet101Hparams, UnetHparams)
from composer.optim import (AdamHparams, AdamWHparams, DecoupledAdamWHparams, DecoupledSGDWHparams, OptimizerHparams,
                            RAdamHparams, RMSPropHparams, SchedulerHparams, SGDHparams, scheduler)
from composer.trainer.ddp import DDPHparams
from composer.trainer.devices import CPUDeviceHparams, DeviceHparams, GPUDeviceHparams

if TYPE_CHECKING:
    from composer.trainer.trainer import Trainer

optimizer_registry = {
    'adam': AdamHparams,
    'adamw': AdamWHparams,
    'decoupled_adamw': DecoupledAdamWHparams,
    'radam': RAdamHparams,
    'sgd': SGDHparams,
    'decoupled_sgdw': DecoupledSGDWHparams,
    'rmsprop': RMSPropHparams,
}

scheduler_registry = {
    'step': scheduler.StepLRHparams,
    'multistep': scheduler.MultiStepLRHparams,
    'exponential': scheduler.ExponentialLRHparams,
    'cosine_decay': scheduler.CosineAnnealingLRHparams,
    'cosine_warmrestart': scheduler.CosineAnnealingWarmRestartsHparams,
    'warmup': scheduler.WarmUpLRHparams,
    'constant': scheduler.ConstantLRHparams,
}

model_registry = {
    "unet": UnetHparams,
    "efficientnetb0": EfficientNetB0Hparams,
    "resnet56_cifar10": CIFARResNetHparams,
    "resnet101": ResNet101Hparams,
    "resnet50": ResNet50Hparams,
    "resnet18": ResNet18Hparams,
    "mnist_classifier": MnistClassifierHparams,
    "gpt2": GPT2Hparams,
}

dataset_registry = {
    "brats": datasets.BratsDatasetHparams,
    "imagenet": datasets.ImagenetDatasetHparams,
    "cifar10": datasets.CIFAR10DatasetHparams,
    "synthetic": datasets.SyntheticDatasetHparams,
    "mnist": datasets.MNISTDatasetHparams,
    "lm": datasets.LMDatasetHparams,
}

algorithms_registry = get_algorithm_registry()

callback_registry = {
    "pytorch_profiler": TorchProfilerHparams,
    "speed_monitor": SpeedMonitorHparams,
    "benchmarker": BenchmarkerHparams,
    "lr_monitor": LRMonitorHparams,
    "grad_monitor": GradMonitorHparams,
}

device_registry = {
    "gpu": GPUDeviceHparams,
    "cpu": CPUDeviceHparams,
}


@dataclass
class TrainerHparams(hp.Hparams):
    """Params for the :class:`Trainer`.

    See the documentation for the :class:`Trainer`.
    """
    hparams_registry = {  # type: ignore
        "algorithms": algorithms_registry,
        "optimizer": optimizer_registry,
        "schedulers": scheduler_registry,
        "loggers": logger_registry,
        "model": model_registry,
        "train_dataset": dataset_registry,
        "val_dataset": dataset_registry,
        "callbacks": callback_registry,
        "device": device_registry,
    }

    device: DeviceHparams = hp.required(doc="Device Parameters")
    train_dataset: datasets.DatasetHparams = hp.required(doc="Training dataset hparams")
    val_dataset: datasets.DatasetHparams = hp.required(doc="Validation dataset hparams")

    optimizer: OptimizerHparams = hp.required(doc="Optimizer to use")

    model: ModelHparams = hp.required(doc="model")
    loggers: List[BaseLoggerBackendHparams] = hp.required(doc="loggers to use")

    max_epochs: int = hp.required(
        doc="training time in epochs and/or batches (e.g., 90ep5ba)",
        template_default=10,
    )

    total_batch_size: int = hp.required(
        doc="batch size for each optimization step, across all devices and gradient accumulations.",
        template_default=2048,
    )
    eval_batch_size: int = hp.required(
        doc="batch size to use for each evaluation step",
        template_default=2048,
    )

    dataloader: DataloaderHparams = hp.required(doc="dataloader hparams")

    grad_accum: int = hp.required(
        template_default=1,
        doc=
        "Determines the number of microbatches to split a per-gpu batch into, used to compensate for low-memory-capacity devices."
    )
    precision: Precision = hp.required(doc="Precision to use for training", template_default=Precision.AMP)
    ddp: DDPHparams = hp.optional(doc="DDP configuration", default_factory=DDPHparams)

    grad_clip_norm: Optional[float] = hp.optional(
        default=None, doc='the norm to clip gradient magnitudes to. Default: None (no clip)')

    algorithms: List[AlgorithmHparams] = hp.optional(doc="Algorithms to employ", default_factory=list)
    schedulers: List[SchedulerHparams] = hp.optional(doc="Scheduler sequence", default_factory=list)
    seed: Optional[int] = hp.optional(default=None, doc="random seed to set")
    validate_every_n_epochs: int = hp.optional(
        doc="Validate every N epochs. Set to -1 to never validate on a epochwise frequency. Defaults to 1", default=1)
    validate_every_n_batches: int = hp.optional(
        doc="Validate every N batches. Set to -1 to never validate on a batchwise frequency. Defaults to -1.",
        default=-1)
    callbacks: List[CallbackHparams] = hp.optional(doc="Callback hparams", default_factory=list)

    checkpoint_filepath: Optional[str] = hp.optional(doc="Path to an existing checkpoint file to load from.",
                                                     default=None)

    checkpoint_interval_unit: Optional[str] = hp.optional(
        doc=
        "Unit for the checkpoint save interval -- should be 'ep' for epochs; 'ba' for batches, or None to disable checkpointing",
        default=None)
    checkpoint_interval: int = hp.optional(doc="Interval for checkpointing.", default=1)
    checkpoint_folder: str = hp.optional(doc="Folder in which to save checkpoint files", default="checkpoints")
    deterministic_mode: bool = hp.optional(doc="Run the model deterministically. Experimental. Performance"
                                           "degradations expected. Certain Torch modules may not have"
                                           "deterministic implementations, which will result in a crash.",
                                           default=False)

    compute_training_metrics: bool = hp.optional(doc="Log validation metrics on training data", default=False)
    log_level: str = hp.optional(doc="Python loglevel to use composer", default="INFO")

    def validate(self):
        super().validate()

        num_procs = 1
        if isinstance(self.device, GPUDeviceHparams) and self.device.n_gpus > 0:
            num_procs = self.device.n_gpus
        if isinstance(self.device, CPUDeviceHparams) and self.device.n_cpus > 0:
            num_procs = self.device.n_cpus

        if self.total_batch_size % (num_procs * self.ddp.num_nodes) != 0:
            raise ValueError(
                f"batch size ({self.total_batch_size}) not divisible by the number of proccesses per node ({num_procs}) "
                f"times the number of nodes ({self.ddp.num_nodes} ")

        if self.eval_batch_size % (num_procs * self.ddp.num_nodes) != 0:
            raise ValueError(
                f"eval batch size ({self.eval_batch_size}) not divisible by the number of proccesses per node ({num_procs}) "
                f"times the number of nodes ({self.ddp.num_nodes}")

    def initialize_object(self) -> Trainer:
        from composer.trainer.trainer import Trainer
        return Trainer.create_from_hparams(hparams=self)

    @classmethod
    def load(cls, model: str) -> TrainerHparams:
        model_hparams_file = os.path.join(
            os.path.dirname(composer.__file__),
            "yamls",
            "models",
            f"{model}.yaml",
        )
        trainer_hparams = TrainerHparams.create(model_hparams_file)
        assert isinstance(trainer_hparams, TrainerHparams), "trainer hparams should return an instance of self"
        return trainer_hparams



# Copyright 2021 MosaicML. All Rights Reserved.

import glob
import os

import pytest

import composer
import composer.algorithms as algorithms
import composer.trainer as trainer
from composer.algorithms.scale_schedule.scale_schedule import ScaleScheduleHparams
from composer.core.precision import Precision
from composer.datasets import SyntheticDatasetHparams
from composer.trainer.devices.device_hparams import CPUDeviceHparams

# skip loading models from huggingface, and resnet101 timeout
EXCLUDE_MODELS = ['gpt2_114m', 'gpt2_85m', 'gpt2_38m']

modeldir_path = os.path.join(os.path.dirname(composer.__file__), 'yamls', 'models')
model_names = glob.glob(os.path.join(modeldir_path, '*.yaml'))
model_names = [os.path.basename(os.path.splitext(mn)[0]) for mn in model_names]
model_names = [name for name in model_names if name not in EXCLUDE_MODELS]


@pytest.mark.parametrize('model_name', model_names)
def test_load(model_name: str, dummy_dataset_hparams: SyntheticDatasetHparams):
    if "gpt" in model_name:
        pytest.skip("GPT doesn't work on the no-op model class")

    dummy_dataset_hparams.sample_pool_size = 4096

    trainer_hparams = trainer.load(model_name)
    trainer_hparams.precision = Precision.FP32
    algs = algorithms.list_algorithms()
    trainer_hparams.algorithms = algorithms.load_multiple(*algs)
    trainer_hparams.train_dataset = dummy_dataset_hparams
    trainer_hparams.val_dataset = dummy_dataset_hparams
    trainer_hparams.device = CPUDeviceHparams(1)
    my_trainer = trainer_hparams.initialize_object()

    assert isinstance(my_trainer, trainer.Trainer)


@pytest.mark.parametrize("ssr", ["0.25", "0.33", "0.50", "0.67", "0.75", "1.00", "1.25"])
def test_scale_schedule_load(ssr: str, dummy_dataset_hparams: SyntheticDatasetHparams):
    dummy_dataset_hparams.sample_pool_size = 4096
    trainer_hparams = trainer.load("classify_mnist")
    trainer_hparams.precision = Precision.FP32
    algs = [f"scale_schedule/{ssr}"]
    trainer_hparams.algorithms = algorithms.load_multiple(*algs)
    trainer_hparams.train_dataset = dummy_dataset_hparams
    trainer_hparams.val_dataset = dummy_dataset_hparams
    trainer_hparams.device = CPUDeviceHparams(1)
    assert len(trainer_hparams.algorithms) == 1
    alg = trainer_hparams.algorithms[0]
    assert isinstance(alg, ScaleScheduleHparams)
    assert alg.ratio == float(ssr)
    my_trainer = trainer_hparams.initialize_object()
    assert isinstance(my_trainer, trainer.Trainer)



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import functools
import logging
from dataclasses import asdict, dataclass
from typing import Optional, Type

import torch
import yahp as hp

from composer.algorithms import AlgorithmHparams
from composer.algorithms.stochastic_depth.sample_stochastic_layers import SampleStochasticBottleneck
from composer.algorithms.stochastic_depth.stochastic_layers import StochasticBottleneck
from composer.core import Algorithm, Event, Logger, State, surgery
from composer.models.resnets import Bottleneck

log = logging.getLogger(__name__)

_VALID_LAYER_DISTRIBUTIONS = ("uniform", "linear")

STOCHASTIC_LAYER_MAPPING = {
    'block': {
        'ResNetBottleneck': (Bottleneck, StochasticBottleneck)
    },
    'sample': {
        'ResNetBottleneck': (Bottleneck, SampleStochasticBottleneck)
    }
}


def validate_stochastic_hparams(stochastic_method: str,
                                target_layer_name: str,
                                drop_rate: float,
                                drop_distribution: str,
                                drop_warmup: float = 0.0):
    """Helper function to validate the Stochastic Depth hyperparameter values"""

    if stochastic_method and (stochastic_method not in STOCHASTIC_LAYER_MAPPING):
        raise ValueError(f"stochastic_method {stochastic_method} is not supported."
                         f" Must be one of {list(STOCHASTIC_LAYER_MAPPING.keys())}")

    if target_layer_name and (target_layer_name not in STOCHASTIC_LAYER_MAPPING[stochastic_method]):
        raise ValueError(f"target_layer_name {target_layer_name} is not supported with {stochastic_method}."
                         f" Must be one of {list(STOCHASTIC_LAYER_MAPPING[stochastic_method].keys())}")

    if drop_rate and (drop_rate < 0 or drop_rate > 1):
        raise ValueError(f"drop_rate must be between 0 and 1: {drop_rate}")

    if drop_distribution and (drop_distribution not in _VALID_LAYER_DISTRIBUTIONS):
        raise ValueError(f"drop_distribution '{drop_distribution}' is"
                         f" not supported. Must be one of {list(_VALID_LAYER_DISTRIBUTIONS)}")

    if drop_warmup and (drop_warmup < 0 or drop_warmup > 1):
        raise ValueError(f"drop_warmup must be between 0 and 1: {drop_warmup}")

    if stochastic_method == "sample" and drop_warmup > 0:
        raise ValueError(f"drop_warmup can not be used with 'sample' stochastic_method")


@dataclass
class StochasticDepthHparams(AlgorithmHparams):
    """Hyperparameters for the Stochastic Depth algorithm"""

    stochastic_method: str = hp.required('The version of stochastic depth to use. Either ["sample", "block"].')
    target_layer_name: str = hp.required(
        f'Reference name of layer to replace. "block" method can be {list(STOCHASTIC_LAYER_MAPPING["block"].keys())}.'
        f' "sample" method can be {list(STOCHASTIC_LAYER_MAPPING["sample"].keys())}.')
    drop_rate: float = hp.optional('The probability of dropping a block or sample.', default=0.2)
    drop_distribution: str = hp.optional(
        '"Uniform" keeps the drop rate the same across blocks. "linear" linearly'
        ' increases the drop rate with block depth until it reaches `drop_rate`.',
        default='linear')
    use_same_gpu_seed: bool = hp.optional(
        'Whether or not to drop the same blocks across GPUs. Only used with "block" method.', default=True)
    drop_warmup: float = hp.optional(
        'Percentage of training to warmup `drop_rate`. Only use with "block" stochastic method.', default=0.0)

    def initialize_object(self) -> StochasticDepth:
        return StochasticDepth(**asdict(self))

    def validate(self):
        super().validate()
        validate_stochastic_hparams(stochastic_method=self.stochastic_method,
                                    target_layer_name=self.target_layer_name,
                                    drop_rate=self.drop_rate,
                                    drop_distribution=self.drop_distribution,
                                    drop_warmup=self.drop_warmup)


def apply_stochastic_depth(model: torch.nn.Module,
                           stochastic_method: str,
                           target_layer_name: str,
                           drop_rate: float = 0.2,
                           drop_distribution: str = 'linear',
                           use_same_gpu_seed: bool = True) -> None:
    """Applies Stochastic Depth algorithm to the specified model.

    The algorithm replaces the specified target layer with a stochastic version
    of the layer. The stochastic layer will randomly drop either samples or the
    layer itself depending on the stochastic method specified. The layer-wise
    version follows the original paper <https://arxiv.org/abs/1603.09382>`_. The
    sample-wise version follows the implementation used for EfficientNet in the
    Tensorflow/TPU repo: <https://github.com/tensorflow/tpu>`_.

    Args:
        stochastic_method: The version of stochastic depth to use. "block" randomly
            drops blocks during training. "sample" randomly drops samples within
            a block during training.
        target_layer_name (str): Block to replace with a stochastic block
            equivalent. The name must be registered in `STOCHASTIC_LAYER_MAPPING` 
            dictionary with the target layer class and the stochastic layer class.
            Currently, must be one of ['ResNetBottleneck'].
        drop_rate (float): The base probability of dropping a layer or sample. Must be
            between 0.0 and 1.0.
        drop_distribution (str): How `drop_rate` is distributed across 
            layers. Value must be one of 'uniform' or 'linear'. 'uniform' assigns
            the same `drop_rate` across all layers. 'linear' linearly 
            increases the drop rate across layer depth starting with 0 drop rate
            and ending with `drop_rate`.
        use_same_gpu_seed (bool): Set to true to have the same layers dropped
            across GPUs when using multi-GPU training. Set to false to have each
            GPU drop a different set of layers. Only used with "block" stochastic
            method.
    """
    validate_stochastic_hparams(stochastic_method=stochastic_method,
                                target_layer_name=target_layer_name,
                                drop_rate=drop_rate,
                                drop_distribution=drop_distribution)
    transforms = {}
    target_layer, stochastic_layer = STOCHASTIC_LAYER_MAPPING[stochastic_method][target_layer_name]
    module_count = surgery.count_module_instances(model, target_layer)
    shared_kwargs = {'drop_rate': drop_rate, 'drop_distribution': drop_distribution, 'module_count': module_count}
    if stochastic_method == 'block':
        rand_generator = torch.Generator()  # Random number generator for each layer
        stochastic_from_target_layer = functools.partial(stochastic_layer.from_target_layer,
                                                         **shared_kwargs,
                                                         use_same_gpu_seed=use_same_gpu_seed,
                                                         rand_generator=rand_generator)
    elif stochastic_method == 'sample':
        stochastic_from_target_layer = functools.partial(stochastic_layer.from_target_layer, **shared_kwargs)
    else:
        raise ValueError(f"stochastic_method {stochastic_method} is not supported."
                         f" Must be one of {list(STOCHASTIC_LAYER_MAPPING.keys())}")
    transforms[target_layer] = stochastic_from_target_layer
    surgery.replace_module_classes(model, policies=transforms)


def _update_drop_rate(module: torch.nn.Module, stochastic_block: Type[torch.nn.Module], drop_rate: float,
                      drop_distribution: str):
    """Recursively updates a module's drop_rate attributes with a new value"""

    if (len(list(module.children())) == 0 and len(list(module.parameters())) > 0):
        return
    else:
        for child in module.children():
            if isinstance(child, stochastic_block):
                if drop_distribution == 'uniform':
                    current_drop_rate = drop_rate
                elif drop_distribution == 'linear':
                    current_drop_rate = ((child.module_id + 1) / child.module_count) * drop_rate  # type: ignore
                else:
                    raise ValueError(f"drop_distribution '{drop_distribution}' is"
                                     f" not supported. Must be one of {list(_VALID_LAYER_DISTRIBUTIONS)}")
                child.drop_rate = torch.tensor(current_drop_rate)
            _update_drop_rate(child, stochastic_block, drop_rate, drop_distribution)


class StochasticDepth(Algorithm):
    """Algorithm to replace a specified block with a stochastic version of the block.

    The stochastic block will randomly drop either samples or the layer itself
    depending on the stochastic method specified. The layer-wise version follows
    the original paper <https://arxiv.org/abs/1603.09382>`_. The sample-wise
    version follows the implementation used for EfficientNet in the Tensorflow/TPU
    repo: <https://github.com/tensorflow/tpu>`_.

    Args:
        stochastic_method: The version of stochastic depth to use. "block" randomly
            drops blocks during training. "sample" randomly drops samples within
            a block during training.
        target_layer_name (str): Which block to replace with a stochastic block
            equivalent. The name must be registered in `STOCHASTIC_LAYER_MAPPING`
            dictionary with the target layer class and the stochastic layer class.
            Currently, must be one of ['ResNetBottleneck'].
        drop_rate (float): The base probability of dropping a layer or a
            sample. Must be between 0.0 and 1.0.
        drop_distribution (str): How `drop_rate` is distributed across
            layers. Value must be either 'uniform' or 'linear'. 'uniform' assigns
            the same `drop_rate` across all layers. 'linear' linearly
            increases the drop rate across layer depth starting with 0 drop rate
            and ending with `drop_rate`.
        drop_warmup (float): Percentage of training epochs to linearly
            increase the drop probability to `linear_drop_rate`. Must be between
            0.0 and 1.0.
        use_same_gpu_seed (bool): Set to true to have the same layers dropped
            across GPUs when using multi-GPU training. Set false to have each
            GPU drop a different set of layers. Only used with "block" stochastic
            method.
        
    """

    def __init__(self,
                 stochastic_method: str,
                 target_layer_name: str,
                 drop_rate: float = 0.2,
                 drop_distribution: str = 'linear',
                 drop_warmup: float = 0.0,
                 use_same_gpu_seed: bool = True):

        if drop_rate == 0.0:
            log.warning('Stochastic Depth will have no effect when drop_rate set to 0')

        if stochastic_method == "sample" and not use_same_gpu_seed:
            log.warning('use_same_gpu_seed=false has no effect when using the "sample" method')

        self.hparams = StochasticDepthHparams(stochastic_method=stochastic_method,
                                              target_layer_name=target_layer_name,
                                              drop_rate=drop_rate,
                                              drop_distribution=drop_distribution,
                                              drop_warmup=drop_warmup,
                                              use_same_gpu_seed=use_same_gpu_seed)

    @property
    def find_unused_parameters(self) -> bool:
        """DDP parameter to notify that parameters may not have gradients if it is dropped during the forward pass."""

        return (self.hparams.stochastic_method == "block")

    def match(self, event: Event, state: State) -> bool:
        """Apply on Event.INIT and Event.BATCH_START if drop_warmup is > 0.0"""

        return (event == Event.INIT) or (event == Event.BATCH_START and self.hparams.drop_warmup > 0.0)

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        assert state.model is not None
        target_layer, stochastic_layer = STOCHASTIC_LAYER_MAPPING[self.hparams.stochastic_method][
            self.hparams.target_layer_name]

        if event == Event.INIT:
            if surgery.count_module_instances(state.model, target_layer) == 0:
                log.warning(f'No {self.hparams.target_layer_name} found in model! Algorithm will function as a no-op.')

            apply_stochastic_depth(state.model,
                                   stochastic_method=self.hparams.stochastic_method,
                                   target_layer_name=self.hparams.target_layer_name,
                                   drop_rate=self.hparams.drop_rate,
                                   drop_distribution=self.hparams.drop_distribution,
                                   use_same_gpu_seed=self.hparams.use_same_gpu_seed)
            num_stochastic_layers = surgery.count_module_instances(state.model, stochastic_layer)
            logger.metric_epoch({'stochastic_depth/num_stochastic_layers': num_stochastic_layers})

        elif event == Event.BATCH_START:
            drop_warmup_iters = state.steps_per_epoch * state.max_epochs * self.hparams.drop_warmup
            if state.step < drop_warmup_iters:
                current_drop_rate = (state.step / drop_warmup_iters) * self.hparams.drop_rate
                _update_drop_rate(state.model, stochastic_layer, current_drop_rate, self.hparams.drop_distribution)
            else:
                current_drop_rate = self.hparams.drop_rate
            logger.metric_batch({'stochastic_depth/drop_rate': current_drop_rate})



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.algorithm_hparams import AlgorithmHparams as AlgorithmHparams
from composer.algorithms.algorithm_registry import get_algorithm_registry as get_algorithm_registry
from composer.algorithms.algorithm_registry import list_algorithms as list_algorithms
from composer.algorithms.alibi import AlibiHparams as AlibiHparams
from composer.algorithms.augmix import AugMix as AugMix
from composer.algorithms.augmix import AugMixHparams as AugMixHparams
from composer.algorithms.blurpool import BlurPool as BlurPool
from composer.algorithms.blurpool import BlurPoolHparams as BlurPoolHparams
from composer.algorithms.channels_last import ChannelsLast as ChannelsLast
from composer.algorithms.channels_last import ChannelsLastHparams as ChannelsLastHparams
from composer.algorithms.colout import ColOut as ColOut
from composer.algorithms.colout import ColOutHparams as ColOutHparams
from composer.algorithms.curriculum_learning import CurriculumLearning as CurriculumLearning
from composer.algorithms.curriculum_learning import CurriculumLearningHparams as CurriculumLearningHparams
from composer.algorithms.cutout import CutOut as CutOut
from composer.algorithms.cutout import CutOutHparams as CutOutHparams
from composer.algorithms.dummy import Dummy as Dummy
from composer.algorithms.dummy import DummyHparams as DummyHparams
from composer.algorithms.ghost_batchnorm import GhostBatchNorm as GhostBatchNorm
from composer.algorithms.ghost_batchnorm import GhostBatchNormHparams as GhostBatchNormHparams
from composer.algorithms.label_smoothing import LabelSmoothing as LabelSmoothing
from composer.algorithms.label_smoothing import LabelSmoothingHparams as LabelSmoothingHparams
from composer.algorithms.layer_freezing import LayerFreezing as LayerFreezing
from composer.algorithms.layer_freezing import LayerFreezingHparams as LayerFreezingHparams
from composer.algorithms.mixup import MixUp as MixUp
from composer.algorithms.mixup import MixUpHparams as MixUpHparams
from composer.algorithms.no_op_model import NoOpModel as NoOpModel
from composer.algorithms.no_op_model import NoOpModelHparams as NoOpModelHparams
from composer.algorithms.progressive_resizing import ProgressiveResizing as ProgressiveResizing
from composer.algorithms.progressive_resizing import ProgressiveResizingHparams as ProgressiveResizingHparams
from composer.algorithms.randaugment import RandAugment as RandAugment
from composer.algorithms.randaugment import RandAugmentHparams as RandAugmentHparams
from composer.algorithms.sam import SAM as SAM
from composer.algorithms.sam import SAMHparams as SAMHparams
from composer.algorithms.scale_schedule import ScaleSchedule as ScaleSchedule
from composer.algorithms.scale_schedule import ScaleScheduleHparams as ScaleScheduleHparams
from composer.algorithms.selective_backprop import SelectiveBackprop as SelectiveBackprop
from composer.algorithms.selective_backprop import SelectiveBackpropHparams as SelectiveBackpropHparams
from composer.algorithms.squeeze_excite import SqueezeExcite as SqueezeExcite
from composer.algorithms.squeeze_excite import SqueezeExcite2d as SqueezeExcite2d
from composer.algorithms.squeeze_excite import SqueezeExciteConv2d as SqueezeExciteConv2d
from composer.algorithms.squeeze_excite import SqueezeExciteHparams as SqueezeExciteHparams
from composer.algorithms.stochastic_depth import StochasticDepth as StochasticDepth
from composer.algorithms.stochastic_depth import StochasticDepthHparams as StochasticDepthHparams
from composer.algorithms.swa import SWA as SWA
from composer.algorithms.swa.hparams import SWAHparams as SWAHparams

load_multiple = AlgorithmHparams.load_multiple
load = AlgorithmHparams.load



# Copyright 2021 MosaicML. All Rights Reserved.

from typing import Optional, Sequence

import torch
import torch.nn.functional as F
from torch import optim

from composer import Algorithm, Engine, Event, Logger, State
from composer.core.types import DataLoader, Optimizer, Precision
from tests.utils.model import SimpleModel


def _get_state(train_dataloader: DataLoader, eval_dataloader: DataLoader, steps_per_epoch: int = 1):
    model = SimpleModel()
    steps_per_epoch = steps_per_epoch
    return State(
        model=model,
        optimizers=optim.SGD(model.parameters(), lr=.001, momentum=0.0),
        max_epochs=1,
        train_batch_size=2,
        eval_batch_size=2,
        train_dataloader=train_dataloader,
        eval_dataloader=eval_dataloader,
        grad_accum=1,
        precision=Precision.FP32,
    )


def test_classifier_trains(
    train_dataloader: DataLoader,
    eval_dataloader: DataLoader,
    algorithms: Sequence[Algorithm],
    n_steps: int = 1,
    check_steps: Optional[Sequence[int]] = None,
):
    if check_steps is None:
        check_steps = list(range(n_steps))

    state = _get_state(train_dataloader=train_dataloader, eval_dataloader=eval_dataloader, steps_per_epoch=n_steps)
    model = state.model

    logger = Logger(state=state, backends=[])
    engine = Engine(state=state, algorithms=algorithms, logger=logger)

    engine.run_event(Event.INIT)
    engine.run_event(Event.TRAINING_START)
    engine.run_event(Event.EPOCH_START)

    def forward(model, X, y, state):
        engine.run_event(Event.BEFORE_FORWARD)
        state.last_input = X
        state.last_target = y
        state.outputs = model(X)
        engine.run_event(Event.AFTER_FORWARD)

    def f_loss(state):
        loss = F.cross_entropy(input=state.outputs, target=state.last_target)
        assert not torch.isnan(loss)
        return loss

    opt = state.optimizers
    assert isinstance(opt, Optimizer)
    assert state.train_dataloader is not None, "train dataloader should be set"

    for step, (X, y) in enumerate(state.train_dataloader):
        # reseed here so data is same for different sets of algorithms
        torch.manual_seed(step)

        # 1st part: run one training step
        engine.run_event(Event.BATCH_START)
        engine.run_event(Event.BEFORE_TRAIN_BATCH)
        forward(model, X, y, state)
        engine.run_event(Event.BEFORE_LOSS)
        state.loss = f_loss(state)
        original_loss = state.loss
        engine.run_event(Event.AFTER_LOSS)
        engine.run_event(Event.BEFORE_BACKWARD)
        state.loss.backward()
        engine.run_event(Event.AFTER_BACKWARD)
        engine.run_event(Event.AFTER_TRAIN_BATCH)
        opt.step()
        opt.zero_grad()
        engine.run_event(Event.BATCH_END)

        # 2nd part; use same data and make sure loss went down
        if step not in check_steps:
            continue
        engine.run_event(Event.BATCH_START)
        engine.run_event(Event.BEFORE_FORWARD)
        state.outputs = model(X)
        engine.run_event(Event.AFTER_FORWARD)
        engine.run_event(Event.BEFORE_LOSS)
        new_loss = f_loss(state)
        assert not torch.isnan(new_loss)
        engine.run_event(Event.BATCH_END)

        # since we used an extremely small learning rate, the loss should
        # be lower on the same input after the update step
        assert new_loss <= original_loss + 1e-20

        # sanity check accuracy computation
        predictions = torch.argmax(state.outputs, dim=1)
        acc = (predictions == y).float().mean()
        assert not torch.any(torch.isnan(acc))



# Copyright 2021 MosaicML. All Rights Reserved.

# To keep the typing organized for this file, see iter_helpers.pyi
# All typing annotations are in there
# All methods signatures must be defined in there.


def map_collection(collection, map_fn):
    """map takes a single element, or a collection of elements, and applies `map_fn` on the element (or each
    element when `maybe_tuple` is a collection).
    It returns the result of `map_fn` in the same data format as `collection` -- i.e. dicts are returned as dicts

    Args:
        collection: The element, or a tuple of elements
        map_fn: A function to invoke on each element

    Returns:
        When `collection` is a single element, it returns the result of `map_fn` on `maybe_tuple`.
        When `collection` is a collection, it returns the results of `map_fn` for each element in
        `collection` in the same data format as the original `collection`
    """
    if isinstance(collection, (tuple, list)):
        return type(collection)(map_fn(x) for x in collection)
    if isinstance(collection, dict):
        return {k: map_fn(v) for k, v in collection.items()}
    return map_fn(collection)


def ensure_tuple(x):
    if isinstance(x, tuple):
        return x
    if isinstance(x, list):
        return tuple(x)
    if isinstance(x, dict):
        return tuple(x.values())
    return (x,)


def zip_collection(item, *others):
    if isinstance(item, (tuple, list, dict)):
        others = [ensure_tuple(other) for other in others]
    else:
        item = [item]
        others = [[other] for other in others]
    return zip(item, *others)



# Copyright 2021 MosaicML. All Rights Reserved.

import os
from typing import Callable, Dict, Type

import pytest

from composer.datasets.brats import BratsDatasetHparams
from composer.datasets.cifar10 import CIFAR10DatasetHparams
from composer.datasets.hparams import DataloaderSpec, DatasetHparams
from composer.datasets.imagenet import ImagenetDatasetHparams
from composer.datasets.lm_datasets import LMDatasetHparams
from composer.datasets.mnist import MNISTDatasetHparams
from composer.datasets.synthetic import SyntheticDatasetHparams
from composer.trainer.trainer_hparams import dataset_registry

ROOT_DATADIR = os.environ.get('MOSAICML_DATASET_DIR', "/tmp")

# for testing, we provide values for required hparams fields
# to initialize test hparams objects
default_required_fields: Dict[Type[DatasetHparams], Callable[[], DatasetHparams]] = {
    #  hparams with empty dicts have no required fields
    CIFAR10DatasetHparams:
        lambda: CIFAR10DatasetHparams(
            is_train=False,
            datadir=os.path.join(ROOT_DATADIR, "cifar10"),
            download=False,
        ),
    BratsDatasetHparams:
        lambda: BratsDatasetHparams(
            is_train=False,
            datadir=os.path.join(ROOT_DATADIR, "01_2d"),
            download=False,
        ),
    ImagenetDatasetHparams:
        lambda: ImagenetDatasetHparams(
            is_train=False,
            datadir=os.path.join(ROOT_DATADIR, "imagenet"),
            crop_size=224,
            resize_size=-1,
        ),
    MNISTDatasetHparams:
        lambda: MNISTDatasetHparams(
            datadir=os.path.join(ROOT_DATADIR, "mnist"),
            is_train=False,
            download=False,
        ),
    SyntheticDatasetHparams:
        lambda: SyntheticDatasetHparams(
            num_classes=100,
            shape=[256, 256],
            one_hot=False,
            sample_pool_size=20,
            device="cpu",
        ),
    LMDatasetHparams:
        lambda: LMDatasetHparams(
            datadir=[os.path.join(ROOT_DATADIR, 'openwebtext_saved')],
            split='train',
            tokenizer_name='gpt2',
        )
}


@pytest.mark.parametrize("dataset_name", dataset_registry.keys())
def test_dataset(dataset_name: str, request: pytest.FixtureRequest) -> None:
    hparams_cls = dataset_registry[dataset_name]
    hparams = default_required_fields[hparams_cls]()
    if dataset_name != "synthetic":
        request.applymarker(pytest.mark.xfail())
    dataloader_spec = hparams.initialize_object()
    assert isinstance(dataloader_spec, DataloaderSpec)



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.utils.string_enum import StringEnum


class Event(StringEnum):
    """An event that occurs during the execution of the training and evaluation loops.
    
    For a conceptual understanding of what events are run within the trainer, see the below *pseudo-code* outline:

    ```
    model = your_model()
    <INIT>  # model surgery here
    optimizers = SGD(model.parameters(), lr=0.01)
    schedulers = CosineAnnealing(optimizers, T_max=90)

    ddp.launch()  # for multi-GPUs, processes are forked here
    <TRAINING_START>  # has access to process rank for DDP

    for epoch in range(90):
        <EPOCH_START>

        for batch in dataloader:
            <AFTER_DATALOADER>
            <BATCH_START>

            #-- closure: forward/backward/loss -- #
            <BEFORE_TRAIN_BATCH>

            # for gradient accumulation
            for microbatch in batch:
                <BEFORE_FORWARD>
                outputs = model.forward(microbatch)
                <AFTER_FORWARD>
                <BEFORE_LOSS>
                loss = model.loss(outputs, microbatch)
                <AFTER_LOSS>
                <BEFORE_BACKWARD>
                loss.backward()
                <AFTER_BACKWARD>

            gradient_unscaling()  # for mixed precision
            gradient_clipping()
            <AFTER_TRAIN_BATCH>
            # -------------------------- #

            optimizer.step() # grad scaling (AMP) also

            <BATCH_END>
            scheduler.step('step')
            maybe_eval()

        scheduler.step('epoch')
        maybe_eval()
        <EPOCH_END>

    <TRAINING_END>

    def maybe_eval():
        <EVAL_START>

        for batch in eval_dataloader:
            <EVAL_BATCH_START>

            <EVAL_BEFORE_FORWARD>
            outputs, targets = model.validate(batch)
            <EVAL_AFTER_FORWARD>

            metrics.update(outputs, targets)
            <EVAL_BATCH_END>
        
        <EVAL_END>
    ```
    """

    INIT = "init"

    TRAINING_START = "training_start"
    EPOCH_START = "epoch_start"
    BATCH_START = "batch_start"

    AFTER_DATALOADER = "after_dataloader"

    BEFORE_TRAIN_BATCH = "before_train_batch"

    BEFORE_FORWARD = "before_forward"
    AFTER_FORWARD = "after_forward"

    BEFORE_LOSS = "before_loss"
    AFTER_LOSS = "after_loss"

    BEFORE_BACKWARD = "before_backward"
    AFTER_BACKWARD = "after_backward"

    AFTER_TRAIN_BATCH = "after_train_batch"

    BATCH_END = "batch_end"
    EPOCH_END = "epoch_end"

    TRAINING_END = "training_end"

    EVAL_START = "eval_start"
    EVAL_BATCH_START = "eval_batch_start"
    EVAL_BEFORE_FORWARD = "eval_before_forward"
    EVAL_AFTER_FORWARD = "eval_after_forward"
    EVAL_BATCH_END = "eval_batch_end"
    EVAL_END = "eval_end"



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import logging
from dataclasses import asdict, dataclass
from typing import Optional

import numpy as np
import torch
import yahp as hp

from composer.algorithms import AlgorithmHparams
from composer.core import Algorithm, Event, Logger, State, surgery

log = logging.getLogger(__name__)

_DEFAULT_GHOST_BATCH_SIZE = 32
_TORCH_BATCHNORM_BASE_CLASS = torch.nn.modules.batchnorm._BatchNorm


def _corresponding_ghost_batchnorm_type(batchnorm: torch.nn.Module):
    if isinstance(batchnorm, torch.nn.BatchNorm1d):
        return GhostBatchNorm1d
    if isinstance(batchnorm, torch.nn.BatchNorm2d):
        return GhostBatchNorm2d
    if isinstance(batchnorm, torch.nn.BatchNorm3d):
        return GhostBatchNorm3d
    raise ValueError(f"Input was of type {type(batchnorm)}, not one of "
                     "torch.nn.BatchNorm1d, torch.nn.BatchNorm2d, torch.nn.BatchNorm3d")


class _GhostBatchNorm(torch.nn.Module):
    """Ghost Batch Normalization. See
    `Train longer, generalize better: closing the generalization gap in large
    batch training of neural networks <https://arxiv.org/abs/1705.08741>`__.

    Works by spliting input into chunks of `ghost_batch_size` samples, and
    running Batch Normalization on each chunk separately. Dim 0 is assumed to
    be the sample axis.

    See also `torch.nn.BatchNorm1d <https://pytorch.org/docs/stable/generated/torch.nn.BatchNorm1d.html>`__,  `torch.nn.BatchNorm2d <https://pytorch.org/docs/stable/generated/torch.nn.BatchNorm2d.html>`__, and
    `torch.nn.BatchNorm3d <https://pytorch.org/docs/stable/generated/torch.nn.BatchNorm3d.html>`__.

    Args:
        ghost_batch_size: the size of the chunks passed into the underlying
            Batch Normalization.
        num_dims: the tensor rank of the inputs to be normalized. Used
            to construct the underlying BatchNorm object when `base_batchnorm`
            is not provided.
        base_batchnorm: A BatchNorm module to be applied to each chunk.

    Raises:
        ValueError: If `ghost_batch_size` exceeds the number of samples in
        the batch provided to `forward`. This might happen when doing
        data-parallel training, because the per-worker batch size is usually
        much smaller than the overall batch size.
    """

    def __init__(self, base_batchnorm: _TORCH_BATCHNORM_BASE_CLASS, ghost_batch_size: int = _DEFAULT_GHOST_BATCH_SIZE):
        super().__init__()
        self.ghost_batch_size = ghost_batch_size
        self.batchnorm = base_batchnorm

    def _has_momentum(self) -> bool:
        return hasattr(self.batchnorm, 'momentum') and self.batchnorm.momentum is not None

    def forward(self, input: torch.Tensor) -> torch.Tensor:  # type: ignore
        batch_size = input.shape[0]
        if batch_size < self.ghost_batch_size:
            raise ValueError(f"Worker batch size {batch_size} < ghost_batch_size {self.ghost_batch_size}")

        nchunks = int(np.ceil(batch_size / self.ghost_batch_size))
        has_momentum = self._has_momentum()
        if has_momentum:
            # applying the same batchnorm multiple times greatly increases
            # the variance of the moving average statistics; reduce the
            # exponential moving average constant proportionally
            # to partially compensate for this
            original_momentum = self.batchnorm.momentum
            self.batchnorm.momentum = float(original_momentum) / nchunks  # type: ignore
        normalized_chunks = [self.batchnorm(chunk) for chunk in input.chunk(nchunks, 0)]
        if has_momentum:
            self.batchnorm.momentum = original_momentum  # type: ignore

        return torch.cat(normalized_chunks, dim=0)

    @staticmethod
    def from_batchnorm(module: torch.nn.Module, ghost_batch_size: int) -> _GhostBatchNorm:
        assert isinstance(module, _TORCH_BATCHNORM_BASE_CLASS), "Module is not a BatchNorm subclass!"
        bn_type = _corresponding_ghost_batchnorm_type(module)
        return bn_type(ghost_batch_size=ghost_batch_size, base_batchnorm=module)


class GhostBatchNorm1d(_GhostBatchNorm):
    pass


class GhostBatchNorm2d(_GhostBatchNorm):
    pass


class GhostBatchNorm3d(_GhostBatchNorm):
    pass


def apply_ghost_batchnorm(model: torch.nn.Module, ghost_batch_size: int) -> torch.nn.Module:
    """Replaces batch normalization modules with ghost batch normalization modules

    This algorithm should be applied before the model has been moved to
    accelerators, and before the model's parameters have been passed to an optimizer.

    Args:
        model: model to transform
        ghost_batch_size: size of sub-batches to normalize over
    """

    def maybe_replace(module: torch.nn.Module, module_index: int) -> Optional[torch.nn.Module]:
        if isinstance(module, _TORCH_BATCHNORM_BASE_CLASS):
            return _GhostBatchNorm.from_batchnorm(module, ghost_batch_size=ghost_batch_size)

    # we have to specify class names explicitly because replace_module_classes
    # now checks if `module.__class__ == cls`, rather than `isinstance(module, cls)`
    transforms = {cls: maybe_replace for cls in [torch.nn.BatchNorm1d, torch.nn.BatchNorm2d, torch.nn.BatchNorm3d]}
    surgery.replace_module_classes(model, policies=transforms)
    return model


@dataclass
class GhostBatchNormHparams(AlgorithmHparams):

    ghost_batch_size: int = hp.required(doc='Size of sub-batches to normalize over',
                                        template_default=_DEFAULT_GHOST_BATCH_SIZE)

    def initialize_object(self) -> "GhostBatchNorm":
        return GhostBatchNorm(**asdict(self))


class GhostBatchNorm(Algorithm):
    """Algorithm to apply `Ghost Batch Normalization <https://arxiv.org/abs/1705.08741>`__ to the model.

    This entails replacing all of the batch normalization modules with ghost
    batch normalization modules on Event.INIT.

    Args:
        ghost_batch_size: size of sub-batches to normalize over
    """

    def __init__(self, ghost_batch_size: int = _DEFAULT_GHOST_BATCH_SIZE):
        self.ghost_batch_size = ghost_batch_size

    def match(self, event: Event, state: State) -> bool:
        """ Runs on Event.INIT
        """
        return event == Event.INIT

    def apply(self, event: Event, state: State, logger: Optional[Logger] = None) -> None:
        """ Applies GhostBatchNorm by wrapping existing BatchNorm modules
        """
        assert state.model is not None, "Model must be in state"

        apply_ghost_batchnorm(model=state.model, ghost_batch_size=self.ghost_batch_size)
        self._log_results(event, state, logger)

    def _log_results(self, event: Event, state: State, logger: Optional[Logger] = None) -> None:
        """Logs the result of GhostBatchNorm applications, including the number
        of modules that have been replaced.
        """
        assert state.model is not None

        num_new_modules = surgery.count_module_instances(state.model, _GhostBatchNorm)
        classname = 'GhostBatchNorm'
        module_name = 'GhostBatchNorm'

        # python logger
        log.info(f'Applied {classname} to model {state.model.__class__.__name__} '
                 f'with ghost_batch_size={self.ghost_batch_size}, '
                 f'Model now has {num_new_modules} {module_name} modules')

        if logger is not None:
            logger.metric_fit({
                f'{classname}/num_new_modules': num_new_modules,
            })
